cimport cython
from libc.math cimport sqrt
from cython.parallel import prange
cdef double PI = 3.14159265359
import numpy as np
cimport numpy as np

@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
@cython.wraparound(False)
cdef class Rbm:
    """
    Rigid body motion (RBM)
    
    ...

    Parameters
    ----------
    radius: float
        Radius of the particles (a)
    particles: int
        Number of particles (Np) 
    viscosity: float  
        Viscosity of the fluid (eta)
    
    """

    def __init__(self, radius=1, particles=1, viscosity=1.0):
        self.a   = radius
        self.Np  = particles
        self.eta = viscosity
        self.mu  = 1.0/(6*PI*self.eta*self.a)
        self.muv = 1.0/(8*PI*self.eta)
        self.mur = 1.0/(8*PI*self.eta*self.a**3)

        self.Mobility = np.zeros( (3*self.Np, 3*self.Np), dtype=np.float64)


    cpdef mobilityTT(self, double [:] v, double [:] r, double [:] F):
        """
        Compute velocity due to body forces using :math:`v=\mu^{TT}\cdot F` 
        ...

        Parameters
        ----------
        v: np.array
            An array of velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
    
        Examples
        --------
        An example of the RBM 

        >>> import pystokes, numpy as np, matplotlib.pyplot as plt
        >>> # particle radius, self-propulsion speed, number and fluid viscosity
        >>> b, vs, Np, eta = 1.0, 1.0, 128, 0.1
        >>> #initialise
        >>> r = pystokes.utils.initialCondition(Np)  # initial random distribution of positions
        >>> p = np.zeros(3*Np); p[2*Np:3*Np] = -1    # initial orientation of the colloids
        >>> 
        >>> rbm = pystokes.wallBounded.Rbm(radius=b, particles=Np, viscosity=eta)
        >>> force = pystokes.forceFields.Forces(particles=Np)
        >>> 
        >>> def rhs(rp):
        >>>     # assign fresh values at each time step
        >>>     r = rp[0:3*Np];   p = rp[3*Np:6*Np]
        >>>     F, v, o = np.zeros(3*Np), np.zeros(3*Np), np.zeros(3*Np)
        >>> 
        >>>     force.lennardJonesWall(F, r, lje=0.01, ljr=5, wlje=1.2, wljr=3.4)
        >>>     rbm.mobilityTT(v, r, F)
        >>>     return np.concatenate( (v,o) )
        >>> 
        >>> # simulate the resulting system
        >>> Tf, Npts = 150, 200
        >>> pystokes.utils.simulate(np.concatenate((r,p)), 
        >>>    Tf,Npts,rhs,integrator='odeint', filename='crystallization')
        """

        cdef int i, j, Np=self.Np, xx=2*Np
        cdef double dx, dy, dz, idr, idr3, idr5, Fdotidr, h2, hsq, tempF
        cdef double vx, vy, vz
        cdef double mu=self.mu, muv=self.muv, a2=self.a*self.a/3.0

        for i in prange(Np, nogil=True):
            vx=0; vy=0; vz=0;
            for j in range(Np):
                dx = r[i]    - r[j]
                dy = r[i+Np]  - r[j+Np]
                h2  =  2*r[j+xx]; hsq=r[j+xx]*r[j+xx]
                if i!=j:
                    dz = r[i+xx]  - r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr3=idr*idr*idr
                    Fdotidr = (F[j] * dx + F[j+Np] * dy + F[j+xx] * dz)*idr*idr
                    #
                    vx += (F[j]   +Fdotidr*dx)*idr + a2*(2*F[j]   -6*Fdotidr*dx)*idr3
                    vy += (F[j+Np]+Fdotidr*dy)*idr + a2*(2*F[j+Np]-6*Fdotidr*dy)*idr3
                    vz += (F[j+xx]+Fdotidr*dz)*idr + a2*(2*F[j+xx]-6*Fdotidr*dz)*idr3

                    ##contributions from the image
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr3 = idr*idr*idr
                    idr5 = idr3*idr*idr
                    Fdotidr = ( F[j]*dx + F[j+Np]*dy + F[j+xx]*dz )*idr*idr

                    vx += -(F[j]   +Fdotidr*dx)*idr - a2*(2*F[j]   -6*Fdotidr*dx)*idr3
                    vy += -(F[j+Np]+Fdotidr*dy)*idr - a2*(2*F[j+Np]-6*Fdotidr*dy)*idr3
                    vz += -(F[j+xx]+Fdotidr*dz)*idr - a2*(2*F[j+xx]-6*Fdotidr*dz)*idr3

                    tempF  = -F[j+xx]     # F_i = M_ij F_j, reflection of the strength
                    Fdotidr = ( F[j]*dx + F[j+Np]*dy + tempF*dz )*idr*idr

                    vx += -h2*(dz*(F[j]   - 3*Fdotidr*dx) + tempF*dx)*idr3
                    vy += -h2*(dz*(F[j+Np]- 3*Fdotidr*dy) + tempF*dy)*idr3
                    vz += -h2*(dz*(tempF  - 3*Fdotidr*dz) + tempF*dz)*idr3 + h2*Fdotidr*idr

                    vx += hsq*( 2*F[j]   - 6*Fdotidr*dx )*idr3
                    vy += hsq*( 2*F[j+Np]- 6*Fdotidr*dy )*idr3
                    vz += hsq*( 2*tempF  - 6*Fdotidr*dz )*idr3

                    vx += 12*a2*dz*( dz*F[j]   - 5*dz*Fdotidr*dx + 2*tempF*dx )*idr5
                    vy += 12*a2*dz*( dz*F[j+Np]- 5*dz*Fdotidr*dy + 2*tempF*dy )*idr5
                    vz += 12*a2*dz*( dz*tempF  - 5*dz*Fdotidr*dz + 2*tempF*dz )*idr5

                    vx += -h2*6*a2*(dz*F[j]   -5*Fdotidr*dx*dz + tempF*dx)*idr5
                    vy += -h2*6*a2*(dz*F[j+Np]-5*Fdotidr*dy*dz+ tempF*dy)*idr5
                    vz += -h2*6*a2*(dz*tempF  -5*Fdotidr*dz*dz + tempF*dz)*idr5 -6*a2*h2*Fdotidr*idr3
                else:
                    ''' self contribution from the image point'''
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/dz
                    idr3 = idr*idr*idr
                    idr5 = idr3*idr*idr
                    Fdotidr = F[j+xx]*dz*idr*idr

                    vx += -(F[j]               )*idr - a2*(2*F[j]    )*idr3
                    vy += -(F[j+Np]            )*idr - a2*(2*F[j+Np] )*idr3
                    vz += -(F[j+xx]+Fdotidr*dz )*idr - a2*(2*F[j+xx] - 6*Fdotidr*dz )*idr3

                    tempF  = -F[j+xx]     # F_i = M_ij F_j, reflection of the strength
                    Fdotidr = tempF*dz*idr*idr

                    vx += -h2*(dz*(F[j]  ) + tempF*dx)*idr3
                    vy += -h2*(dz*(F[j+Np]) + tempF*dy)*idr3
                    vz += -h2*(dz*(tempF -3*Fdotidr*dz) + tempF*dz)*idr3 + h2*Fdotidr*idr

                    vx += hsq*( 2*F[j]   )*idr3
                    vy += hsq*( 2*F[j+Np])*idr3
                    vz += hsq*( 2*tempF  - 6*Fdotidr*dz )*idr3

                    vx += 12*a2*dz*( dz*F[j]   )*idr5
                    vy += 12*a2*dz*( dz*F[j+Np])*idr5
                    vz += 12*a2*dz*( dz*tempF  - 5*dz*Fdotidr*dz + 2*tempF*dz )*idr5

                    vx += -h2*6*a2*(dz*F[j]   )*idr5
                    vy += -h2*6*a2*(dz*F[j+Np])*idr5
                    vz += -h2*6*a2*(dz*tempF  -5*Fdotidr*dz*dz+tempF*dz)*idr5 -6*a2*h2*Fdotidr*idr3

            v[i  ]  += mu*F[i]    + muv*vx
            v[i+Np] += mu*F[i+Np] + muv*vy
            v[i+xx] += mu*F[i+xx] + muv*vz
        return


    cpdef mobilityTR(self, double [:] v, double [:] r, double [:] T):
        """
        Compute velocity due to body torque using :math:`v=\mu^{TR}\cdot T` 
        ...

        Parameters
        ----------
        v: np.array
            An array of velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        T: np.array
            An array of torques
            An array of size 3*Np,
        """

        cdef int Np = self.Np, i, j, xx=2*Np
        cdef double dx, dy, dz, idr, idr3, rlz, Tdotidr, h2,
        cdef double vx, vy, vz,
        cdef double muv=self.muv

        for i in prange(Np, nogil=True):
            vx=0; vy=0; vz=0;
            for j in range(Np):
                dx = r[i]   - r[j]
                dy = r[i+Np] - r[j+Np]
                h2 = 2*r[i+xx]
                if i != j:
                    #contributions from the source
                    dz = r[i+xx] - r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr3 = idr*idr*idr

                    vx += (T[j+Np]*dz - T[j+xx]*dy )*idr3
                    vy += (T[j+xx]*dx - T[j]   *dz )*idr3
                    vz += (T[j]   *dy - T[j+Np]*dx )*idr3

                    #contributions from the image
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr3 = idr*idr*idr

                    vx += -(T[j+Np]*dz - T[j+xx]*dy )*idr3
                    vy += -(T[j+xx]*dx - T[j]   *dz )*idr3
                    vz += -(T[j]   *dy - T[j+Np]*dx )*idr3

                    rlz = (dx*T[j+Np] - dy*T[j])*idr*idr
                    vx += (h2*(T[j+Np]-3*rlz*dx) + 6*dz*dx*rlz)*idr3
                    vy += (h2*(-T[j]  -3*rlz*dy) + 6*dz*dy*rlz)*idr3
                    vz += (h2*(       -3*rlz*dz) + 6*dz*dz*rlz)*idr3
                else:
                    ''' the self contribution from the image point'''
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/dz
                    idr3 = idr*idr*idr

                    vx += -(T[j+Np]*dz )*idr3
                    vy += -(- T[j] *dz )*idr3

                    vx += h2*T[j+Np]*idr3
                    vy += -h2*T[j]*idr3

            v[i]    -= muv*vx
            v[i+Np] -= muv*vy
            v[i+xx] -= muv*vz
        return


    cpdef propulsionT2s(self, double [:] v, double [:] r, double [:] S):
        """
        Compute velocity due to 2s mode of the slip :math:`v=\pi^{T,2s}\cdot S` 
        ...

        Parameters
        ----------
        v: np.array
            An array of velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        S: np.array
            An array of 2s mode of the slip
            An array of size 5*Np,
        """

        cdef int Np=self.Np, i, j, xx=2*Np, xx1=3*Np , xx2=4*Np
        cdef double dx, dy, dz, idr, idr2, idr3, idr5, idr4, aidr2, trS, h2, hsq
        cdef double sxx, syy, szz, sxy, syx, syz, szy, sxz, szx, srr, srx, sry, srz
        cdef double Sljrlx, Sljrly, Sljrlz, Sljrjx, Sljrjy, Sljrjz
        cdef double vx, vy, vz, mus=(28.0*self.a**3)/24

        for i in prange(Np, nogil=True):
            vx=0; vy=0;   vz=0;
            for j in  range(Np):
                h2 = 2*r[j+xx]; hsq = r[j+xx]*r[j+xx];
                sxx = S[j]  ; syy = S[j+Np]; szz = -sxx-syy;
                sxy = S[j+xx]; syx = sxy;
                sxz = S[j+xx1]; szx = sxz;
                syz = S[j+xx2]; szy = syz;
                dx = r[i]   - r[j]
                dy = r[i+Np] - r[j+Np]
                if i!=j:
                    dz = r[i+xx] - r[j+xx]
                    idr  = 1.0/sqrt( dx*dx + dy*dy + dz*dz );
                    idr2 = idr*idr; idr3 = idr2*idr; idr5 = idr3*idr2;
                    srx = (sxx*dx +  sxy*dy + sxz*dz );
                    sry = (sxy*dx +  syy*dy + syz*dz );
                    srz = (sxz*dx +  syz*dy + szz*dz );
                    srr = srx*dx + sry*dy + srz*dz

                    ## contributions from the source
                    vx += 3*srr*dx*idr5;
                    vy += 3*srr*dy*idr5;
                    vz += 3*srr*dz*idr5;

                    ## contributions from the image
                    dz = r[i+xx]+r[j+xx]
                    idr  = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr2 = idr*idr; idr3 = idr2*idr; idr5 = idr3*idr2;

                    
                    #reflecting the first index of stresslet, S_jl M_lm
                    sxz=-sxz; syz=-syz; szz=-szz;     trS=sxx+syy+szz;
                    Sljrlx = sxx*dx +  sxy*dx + sxz*dx ;
                    Sljrly = syx*dy +  syy*dy + syz*dy ;
                    Sljrlz = szx*dz +  szy*dz + szz*dz ;
                    
                    Sljrjx = sxx*dx +  sxy*dy + sxz*dz ;
                    Sljrjy = syx*dx +  syy*dy + syz*dz ;
                    Sljrjz = szx*dx +  szy*dy + szz*dz ;
                    srx = sxx*dx + sxy*dy + sxz*dz
                    sry = syx*dx + syy*dy + syz*dz
                    srz = sxz*dx + syz*dy + szz*dz
                    srr = (srx*dx + sry*dy + srz*dz)*idr2

                    vx += (-Sljrlx + Sljrjx + trS*dx -3*srr*dx)*idr3
                    vy += (-Sljrly + Sljrjy + trS*dy -3*srr*dy)*idr3
                    vz += (-Sljrlz + Sljrjz + trS*dz -3*srr*dz)*idr3

                    vx += -2*(dz*(sxz-3*srz*dx*idr2)+ szz*dx)*idr3;
                    vy += -2*(dz*(syz-3*srz*dy*idr2)+ szz*dy)*idr3;
                    vz += -2*(dz*(szz-3*srz*dz*idr2)+ szz*dz - srz)*idr3;

                    vx += h2*( sxz-3*srz*dx*idr2)*idr3;
                    vy += h2*( syz-3*srz*dy*idr2)*idr3;
                    vz += h2*( szz-3*srz*dz*idr2)*idr3;

                    #reflecting both the indices of stresslet, S_jl M_lm M_jk
                    szx = -szx ; szy = -szy; szz = -szz;
                    srx = (sxx*dx +  sxy*dy + sxz*dz )
                    sry = (sxy*dx +  syy*dy + syz*dz )
                    srz = (sxz*dx +  syz*dy + szz*dz )
                    srr = (srx*dx + sry*dy + srz*dz)*idr2

                    vx += h2*( (dz*(-6*srx + 15*srr*dx)-3*srz*dx)*idr5 + (sxz)*idr3) ;
                    vy += h2*( (dz*(-6*sry + 15*srr*dy)-3*srz*dy)*idr5 + (syz)*idr3) ;
                    vz += h2*( (dz*(-6*srz + 15*srr*dz)-3*srz*dz)*idr5 + (szz + 3*srr)*idr3);

                    vx += hsq*(12*srx - 30*srr*dx)*idr5
                    vy += hsq*(12*sry - 30*srr*dy)*idr5
                    vz += hsq*(12*srz - 30*srr*dz)*idr5

                else:
                    ''' the self contribution from the image point'''
                    dz = r[i+xx]+r[j+xx]
                    idr  = 1.0/dz
                    idr2 = idr*idr; idr3 = idr2*idr; idr4 = idr3*idr

                    #reflecting the first index of stresslet, S_jl M_lm
                    sxz=-sxz; syz=-syz; szz=-szz;     trS=sxx+syy+szz;

                    Sljrlz = szx*dz +  szy*dz + szz*dz ;
                    Sljrjx = sxz*dz
                    Sljrjy = syz*dz
                    Sljrjz = szz*dz

                    vx += Sljrjx*idr3 ;
                    vy += Sljrjy*idr3 ;
                    vz += (-Sljrlz + Sljrjz + trS*dz -3*szz*dz)*idr3 ;

                    vx += -2*dz*sxz*idr3;
                    vy += -2*dz*syz*idr3;
                    vz += -2*dz*(-2*szz)*idr3;

                    vx +=      h2*sxz*idr3;
                    vy +=      h2*syz*idr3;
                    vz += h2*(-2*szz)*idr3;

                    #reflecting both the indices of stresslet, S_jl M_lm M_jk
                    szx = -szx ; syz = -syz; szz = -szz;

                    vx += h2*dz*(-5*sxz)*idr4
                    vy += h2*dz*(-5*syz)*idr4
                    vz += h2*dz*(10*szz)*idr4

                    vx += hsq*12*sxz*idr4
                    vy += hsq*12*syz*idr4
                    vz += hsq*(-18*szz)*idr4

            v[i]    += vx*mus
            v[i+Np] += vy*mus
            v[i+xx] += vz*mus
        return


    cpdef propulsionT3t(self, double [:] v, double [:] r, double [:] D):
        """
        Compute velocity due to 3t mode of the slip :math:`v=\pi^{T,3t}\cdot D` 
        ...

        Parameters
        ----------
        v: np.array
            An array of velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        D: np.array
            An array of 3t mode of the slip
            An array of size 3*Np,
        """

        cdef int Np=self.Np, i, j, xx=2*Np
        cdef double dx, dy, dz, idr, idr3, idr5, Ddotidr, tempD, hsq, h2
        cdef double vx, vy, vz, mud = 3.0*self.a*self.a*self.a/5, muv = -1.0*(self.a**5)/10

        for i in prange(Np, nogil=True):
            vx=0; vy=0; vz=0;
            for j in range(Np):
                dx = r[i]    - r[j]
                dy = r[i+Np]  - r[j+Np]
                h2  =  2*r[j+xx]
                if i!=j:
                    dz = r[i+xx] - r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr3=idr*idr*idr
                    Ddotidr = (D[j]*dx + D[j+Np]*dy + D[j+xx]*dz)*idr*idr
                    #
                    vx += (2*D[j]    - 6*Ddotidr*dx)*idr3
                    vy += (2*D[j+Np] - 6*Ddotidr*dy)*idr3
                    vz += (2*D[j+xx] - 6*Ddotidr*dz)*idr3

                    ##contributions from the image
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr3 = idr*idr*idr
                    idr5 = idr3*idr*idr
                    Ddotidr = (D[j]*dx + D[j+Np]*dy + D[j+xx]*dz)*idr*idr

                    vx += -(2*D[j]    - 6*Ddotidr*dx )*idr3
                    vy += -(2*D[j+Np] - 6*Ddotidr*dy )*idr3
                    vz += -(2*D[j+xx] - 6*Ddotidr*dz )*idr3

                    tempD = -D[j+xx]     # D_i = M_ij D_j, reflection of the strength
                    Ddotidr = ( D[j]*dx + D[j+Np]*dy + tempD*dz )*idr*idr

                    vx += 12*dz*( dz*D[j]   - 5*dz*Ddotidr*dx + 2*tempD*dx )*idr5
                    vy += 12*dz*( dz*D[j+Np]- 5*dz*Ddotidr*dy + 2*tempD*dy )*idr5
                    vz += 12*dz*( dz*tempD  - 5*dz*Ddotidr*dz + 2*tempD*dz )*idr5

                    vx += -6*h2*(dz*D[j]   -5*Ddotidr*dx*dz + tempD*dx)*idr5
                    vy += -6*h2*(dz*D[j+Np]-5*Ddotidr*dy*dz + tempD*dy)*idr5
                    vz += -6*h2*(dz*tempD  -5*Ddotidr*dz*dz + tempD*dz)*idr5 -6*h2*Ddotidr*idr3

                else:
                    ''' self contribution from the image point'''
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/dz
                    idr3 = idr*idr*idr
                    idr5 = idr3*idr*idr
                    Ddotidr = D[j+xx]*dz*idr*idr

                    vx += -(2*D[j]    )*idr3
                    vy += -(2*D[j+Np] )*idr3
                    vz += -(2*D[j+xx] - 6*Ddotidr*dz )*idr3

                    tempD = -D[j+xx]     # D_i = M_ij D_j, reflection of the strength
                    Ddotidr = tempD*dz*idr*idr

                    vx += 12*dz*( dz*D[j]   )*idr5
                    vy += 12*dz*( dz*D[j+Np])*idr5
                    vz += 12*dz*( dz*tempD  - 5*dz*Ddotidr*dz + 2*tempD*dz )*idr5

                    vx += -6*h2*(dz*D[j]   )*idr5
                    vy += -6*h2*(dz*D[j+Np])*idr5
                    vz += -6*h2*(dz*tempD  -5*Ddotidr*dz*dz + tempD*dz)*idr5 -6*h2*Ddotidr*idr3

            v[i  ]  += muv*vx
            v[i+Np] += muv*vy
            v[i+xx] += muv*vz
        return
   


    ## Angular Velocities
    cpdef mobilityRT(self, double [:] o, double [:] r, double [:] F):
        """
        Compute angular velocity due to body forces using :math:`o=\mu^{RT}\cdot F` 
        ...

        Parameters
        ----------
        o: np.array
            An array of angular velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
        """


        cdef int Np = self.Np, i, j, xx=2*Np
        cdef double dx, dy, dz, idr, idr3, rlz, Fdotidr, h2
        cdef double ox, oy, oz, muv=self.muv

        for i in prange(Np, nogil=True):
            ox=0; oy=0; oz=0;
            for j in range(Np):
                dx = r[i]   - r[j]
                dy = r[i+Np] - r[j+Np]
                h2 = 2*r[i+xx]
                if i != j:
                    #contributions from the source
                    dz = r[i+xx] - r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr3 = idr*idr*idr

                    ox += (F[j+Np]*dz - F[j+xx]*dy )*idr3
                    oy += (F[j+xx]*dx - F[j]   *dz )*idr3
                    oz += (F[j]   *dy - F[j+Np]*dx )*idr3

                    #contributions from the image
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr3 = idr*idr*idr
                    rlz = (dx*F[j+Np] - dy*F[j])*idr*idr

                    ox += -(F[j+Np]*dz - F[j+xx]*dy )*idr3
                    oy += -(F[j+xx]*dx - F[j]   *dz )*idr3
                    oz += -(F[j]   *dy - F[j+Np]*dx )*idr3

                    ox += (h2*(F[j+Np]-3*rlz*dx) + 6*dz*dx*rlz)*idr3
                    oy += (h2*(-F[j]  -3*rlz*dy) + 6*dz*dy*rlz)*idr3
                    oz += (h2*(       -3*rlz*dz) + 6*dz*dz*rlz)*idr3

                else:
                    ''' the self contribution from the image point'''
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/dz
                    idr3 = idr*idr*idr


                    ox += (h2*F[j+Np])*idr3
                    oy += (h2*-F[j]  )*idr3

            o[i  ]  += muv*ox
            o[i+Np] += muv*oy
            o[i+xx] += muv*oz
        return


    cpdef mobilityRR(self, double [:] o, double [:] r, double [:] T):
        """
        Compute angular velocity due to body torques using :math:`o=\mu^{RR}\cdot T` 
        ...

        Parameters
        ----------
        o: np.array
            An array of angular velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        T: np.array
            An array of forces
            An array of size 3*Np,
        """

        cdef int Np=self.Np, i, j, xx=2*Np
        cdef double dx, dy, dz, idr, idr3, idr5, Tdotidr, tempT, hsq, h2
        cdef double ox, oy, oz, mur=self.mur, muv=self.muv

        for i in prange(Np, nogil=True):
            ox=0; oy=0; oz=0;
            for j in range(Np):
                dx = r[i]    - r[j]
                dy = r[i+Np]  - r[j+Np]
                h2  =  2*r[j+xx]
                if i!=j:
                    dz = r[i+xx] - r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr3=idr*idr*idr
                    Tdotidr = (T[j]*dx + T[j+Np]*dy + T[j+xx]*dz)*idr*idr
                    #
                    ox += (2*T[j]    - 6*Tdotidr*dx)*idr3
                    oy += (2*T[j+Np] - 6*Tdotidr*dy)*idr3
                    oz += (2*T[j+xx] - 6*Tdotidr*dz)*idr3

                    ##contributions from the image
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr3 = idr*idr*idr
                    idr5 = idr3*idr*idr
                    Tdotidr = (T[j]*dx + T[j+Np]*dy + T[j+xx]*dz)*idr*idr

                    ox += -(2*T[j]    - 6*Tdotidr*dx )*idr3
                    oy += -(2*T[j+Np] - 6*Tdotidr*dy )*idr3
                    oz += -(2*T[j+xx] - 6*Tdotidr*dz )*idr3

                    tempT = -T[j+xx]     # D_i = M_ij D_j, reflection of the strength
                    Tdotidr = ( T[j]*dx + T[j+Np]*dy + tempT*dz )*idr*idr

                    ox += 12*dz*( dz*T[j]   - 5*dz*Tdotidr*dx + 2*tempT*dx )*idr5
                    oy += 12*dz*( dz*T[j+Np]- 5*dz*Tdotidr*dy + 2*tempT*dy )*idr5
                    oz += 12*dz*( dz*tempT  - 5*dz*Tdotidr*dz + 2*tempT*dz )*idr5

                    ox += -6*h2*(dz*T[j]   -5*Tdotidr*dx*dz + tempT*dx)*idr5
                    oy += -6*h2*(dz*T[j+Np]-5*Tdotidr*dy*dz + tempT*dy)*idr5
                    oz += -6*h2*(dz*tempT  -5*Tdotidr*dz*dz + tempT*dz)*idr5 -6*h2*Tdotidr*idr3

                else:

                    ''' self contribution from the image point'''
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/dz
                    idr3 = idr*idr*idr

                    ox += 4*T[j]*idr3
                    oy += 4*T[j+Np]*idr3
                    oz += 16*T[j+xx]*idr3

            o[i  ]  += mur*T[i  ]  - muv*ox
            o[i+Np] += mur*T[i+Np] - muv*oy
            o[i+xx] += mur*T[i+xx] - muv*oz
        return
        
        
    cpdef propulsionR2s(self, double [:] o, double [:] r, double [:] S):
        cdef int Np=self.Np, i, j, xx=2*Np, xx1=3*Np , xx2=4*Np
        cdef double dx, dy, dz, idr, idr2, idr3, idr5, idr7
        cdef double sxx, syy, szz, sxy, syx, syz, szy, sxz, szx, srr, srx, sry, srz
        cdef double Sljrlx, Sljrly, Sljrlz, Sljrjx, Sljrjy, Sljrjz, rlz, smr3, smkrk3
        cdef double ox, oy, oz, mus = (28.0*self.a**3)/24, h

        for i in prange(Np, nogil=True):
            ox=0;   oy=0;   oz=0;
            sxz = S[i+xx1];
            syz = S[i+xx2];
            for j in range(Np):
                sxx = S[j    ]  ;
                syy = S[j+Np ];
                sxy = S[j+xx ];
                sxz = S[j+xx1];
                syz = S[j+xx2];
                if i != j:
                    #syx = sxy;
                    # szx = sxz;
                    # szy = syz;
                    #szz = -sxx-syy;
                    dx = r[i]   - r[j]
                    dy = r[i+Np] - r[j+Np]
                    h=r[j+xx]
                    dz = r[i+xx] - r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr5 = idr*idr*idr*idr*idr
                    srx = sxx*dx +  sxy*dy + sxz*dz
                    sry = sxy*dx +  syy*dy + syz*dz
                    srz = sxz*dx +  syz*dy - (sxx+syy)*dz

                    ox += 3*(sry*dz - srz*dy )*idr5
                    oy += 3*(srz*dx - srx*dz )*idr5
                    oz += 3*(srx*dy - sry*dx )*idr5
                    #
                    ####contributions from the image
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr5=idr*idr*idr*idr*idr
                    srx = sxx*dx +  sxy*dy + sxz*dz
                    sry = sxy*dx +  syy*dy + syz*dz
                    srz = sxz*dx +  syz*dy - (sxx+syy)*dz

                    ox += 3*(-sry*dz + srz*dy )*idr5
                    oy += 3*(-srz*dx + srx*dz )*idr5
                    oz += 3*(-srx*dy + sry*dx )*idr5

                    #rlz = (dx*syz - dy*sxz)*idr*idr
                    #ox += (2*syz  - 6*rlz*dx)*idr3
                    #oy += (-2*sxz - 6*rlz*dy)*idr3
                    #oz += (       - 6*rlz*dz)*idr3
                    #
                    ###reflecting the second index of stresslet, S_jl M_lm
                    #sxz=-sxz; syz=-syz; szz=-szz;
                    #
                    #smr3 = sxz*dy-syz*dx
                    #ox += 6*(dz*(sxx*dy-syx*dx) + smr3*dx)*idr5
                    #ox += 6*(dz*(sxy*dy-syy*dx) + smr3*dy)*idr5
                    #oz += 6*(dz*(sxz*dy-syz*dx) + smr3*dz)*idr5

                    #Sljrjx = sxx*dx +  sxy*dy + sxz*dz ;
                    #Sljrjy = syx*dx +  syy*dy + syz*dz ;
                    #Sljrjz = szx*dx +  szy*dy + szz*dz ;
                    #srr = (sxx*dx*dx + syy*dy*dy + szz*dz*dz +  2*sxy*dx*dy)*idr2 ;
                    #
                    #ox += 2*syz*idr3 - 3*(Sljrjy*dz - Sljrjz*dy)*idr5
                    #oy += 2*szx*idr3 - 3*(Sljrjz*dx - Sljrjx*dz)*idr5
                    #oz +=              3*(Sljrjx*dy - Sljrjy*dx)*idr5
                    #
                    #smkrk3 = 30*(dx*Sljrjy-dy*Sljrjx)*idr5*idr2
                    #ox += (h+dz)*smkrk3*dx
                    #ox += (h+dz)*smkrk3*dy
                    #oz += (h+dz)*smkrk3*dz

                    #ox += 2*syz*idr3  - 3*(Sljrjy*dz - Sljrjz*dy )*idr5
                    #oy += -2*szy*idr3 - 3*(Sljrjz*dx - Sljrjx*dz )*idr5
                    #oz +=             - 3*(Sljrjx*dy - Sljrjy*dx )*idr5
                    #
                    #ox += 6*h*(-Sljrjy + (sxx*dy-syx*dx))*idr5
                    #oy += 6*h*(Sljrjx  + (sxy*dy-syy*dx))*idr5
                    #ox +=                (sxz*dy-syz*dx)*idr5
                else:
                    ### self contributions from the image
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/dz
                    idr3 = idr*idr*idr
                    # idr5 = idr3*idr*idr
                    ox += 3*(-syz)*idr3
                    oy += 3*(sxz )*idr3
                    ##reflecting the second index of stresslet, S_jl M_lm
                    #sxz=-sxz; syz=-syz; szz=-szz;
                    #
                    #Sljrjx = sxz*dz ;
                    #Sljrjy = syz*dz ;
                    #Sljrjz = szz*dz ;
                    #srr = szz*dz*dz*idr2 ;
                    #
                    #ox += 2*syz*idr3 - 3*(Sljrjy*dz)*idr5
                    #oy += 2*szx*idr3 - 3*(- Sljrjx*dz)*idr5
                    #
                    #smkrk3 = 30*(dx*Sljrjy-dy*Sljrjx)*idr5*idr2
                    #oz += (h+dz)*smkrk3*dz

                    #ox += 2*syz*idr3  - 3*(Sljrjy*dz )*idr5
                    #oy += -2*szy*idr3 - 3*(- Sljrjx*dz )*idr5
                    #
                    #ox += 6*h*(-Sljrjy )*idr5
                    #oy += 6*h*(Sljrjx  )*idr5

            o[i]    += ox
            o[i+Np] += oy
            o[i+xx] += oz
        return


    cpdef propulsionR3a(  self, double [:] o, double [:] r, double [:] V):
        '''
        approximation involved
        '''
        cdef int Np = self.Np, i, j, xx=2*Np
        cdef double dx, dy, dz, idr, idr2, idr4, idr5, vxx, vyy, vxy, vxz, vyz, vrr, vrx, vry, vrz
        cdef double ox, oy, oz

        for i in prange(Np, nogil=True):
            ox=0; oy=0; oz=0;
            for j in range(Np):
                vxx = V[j]
                vyy = V[j+Np]
                vxy = V[j+xx]
                vxz = V[j+3*Np]
                vyz = V[j+4*Np]
                if i != j:
                    pass
                    #dx = r[i]      - r[j]
                    #dy = r[i+Np]   - r[j+Np]
                    #dz = r[i+xx] - r[j+xx]
                    #idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    #idr5 = idr*idr*idr*idr*idr
                    #vrr = (vxx*(dx*dx-dz*dz) + vyy*(dy*dy-dz*dz) +  2*vxy*dx*dy + 2*vxz*dx*dz  +  2*vyz*dy*dz)*idr*idr
                    #vrx = vxx*dx +  vxy*dy + vxz*dz
                    #vry = vxy*dx +  vyy*dy + vyz*dz
                    #vrz = vxz*dx +  vyz*dy - (vxx+vyy)*dz

                    #ox +=  5*( 6*vrx- 15*vrr*dx )*idr5
                    #oy +=  5*( 6*vry- 15*vrr*dy )*idr5
                    #oz +=  5*( 6*vrz- 15*vrr*dz )*idr5
                    #
                    ###contribution from the image point
                    #dz = r[i+xx] + r[j+xx]
                    #idr = 1.0/dz
                    #idr5 = idr*idr*idr*idr*idr
                    #vrr = (vxx*(dx*dx-dz*dz) + vyy*(dy*dy-dz*dz) +  2*vxy*dx*dy + 2*vxz*dx*dz  +  2*vyz*dy*dz)*idr*idr
                    #vrx = vxx*dx +  vxy*dy + vxz*dz
                    #vry = vxy*dx +  vyy*dy + vyz*dz
                    #vrz = vxz*dx +  vyz*dy - (vxx+vyy)*dz

                    #ox +=  -5*( 6*vrx- 15*vrr*dx )*idr5
                    #oy +=  -5*( 6*vry- 15*vrr*dy )*idr5
                    #oz +=  -5*( 6*vrz- 15*vrr*dz )*idr5
                else :
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/dz
                    idr4 = idr*idr*idr*idr
                    #vrr = 45*(vxx+vyy)
                    #ox += -5*(6*vxz*dz)*idr4
                    #oy += -5*(6*vyz*dz)*idr4
                    oz += -45*(vxx+vyy)*idr4
            #o[i  ]  += ox
            #o[i+Np] += oy
            o[i+xx] += oz
        return


    cpdef propulsionR4a(self, double [:] o, double [:] r, double [:] M):
        '''
        approximation involved
        '''
        cdef int Np = self.Np, i, j, xx=2*Np
        cdef double ox, oy, oz, dx, dy, dz, idr, idr5, idr7, idr9, muv = 1
        cdef double mrrr, mrrx, mrry, mrrz, mxxx, myyy, mxxy, mxxz, mxyy, mxyz, myyz

        for i in prange(Np, nogil=True):
            ox=0; oy=0; oz=0;
            for j in range(Np):
                mxxx = M[j]
                myyy = M[j+Np]
                mxxy = M[j+2*Np]
                mxxz = M[j+3*Np]
                mxyy = M[j+4*Np]
                mxyz = M[j+5*Np]
                myyz = M[j+6*Np]
                if i != j:
                    pass
                    #dx = r[i]      - r[j]
                    #dy = r[i+Np]   - r[j+Np]
                    #dz = r[i+2*Np] - r[j+2*Np]
                    #idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    #idr7 = idr*idr*idr*idr*idr*idr*idr
                    #idr9 = idr7*idr*idr
                    #
                    #mrrr = mxxx*dx*(dx*dx-3*dz*dz) + 3*mxxy*dy*(dx*dx-dz*dz) + mxxz*dz*(3*dx*dx-dz*dz) +\
                    #   3*mxyy*dx*(dy*dy-dz*dz) + 6*mxyz*dx*dy*dz + myyy*dy*(dy*dy-3*dz*dz) +  myyz*dz*(3*dy*dy-dz*dz)
                    #mrrx = mxxx*(dx*dx-dz*dz) + mxyy*(dy*dy-dz*dz) +  2*mxxy*dx*dy + 2*mxxz*dx*dz  +  2*mxyz*dy*dz
                    #mrry = mxxy*(dx*dx-dz*dz) + myyy*(dy*dy-dz*dz) +  2*mxyy*dx*dy + 2*mxyz*dx*dz  +  2*myyz*dy*dz
                    #mrrz = mxxz*(dx*dx-dz*dz) + myyz*(dy*dy-dz*dz) +  2*mxyz*dx*dy - 2*(mxxx+mxyy)*dx*dz  - 2*(mxxy+myyy)*dy*dz

                    #ox += -21*mrrr*dx*idr9 + 9*mrrx*idr7
                    #oy += -21*mrrr*dy*idr9 + 9*mrry*idr7
                    #oz += -21*mrrr*dz*idr9 + 9*mrrz*idr7
                    #''' self contribution from the image point'''
                    #dz = r[i+xx] + r[j+xx]
                    #idr = 1.0/dz
                    #idr7 = idr*idr*idr*idr*idr*idr*idr
                    #idr9 = idr7*idr*idr
                    #
                    #mrrr = mxxx*dx*(dx*dx-3*dz*dz) + 3*mxxy*dy*(dx*dx-dz*dz) + mxxz*dz*(3*dx*dx-dz*dz) +\
                    #   3*mxyy*dx*(dy*dy-dz*dz) + 6*mxyz*dx*dy*dz + myyy*dy*(dy*dy-3*dz*dz) +  myyz*dz*(3*dy*dy-dz*dz)
                    #mrrx = mxxx*(dx*dx-dz*dz) + mxyy*(dy*dy-dz*dz) +  2*mxxy*dx*dy + 2*mxxz*dx*dz  +  2*mxyz*dy*dz
                    #mrry = mxxy*(dx*dx-dz*dz) + myyy*(dy*dy-dz*dz) +  2*mxyy*dx*dy + 2*mxyz*dx*dz  +  2*myyz*dy*dz
                    #mrrz = mxxz*(dx*dx-dz*dz) + myyz*(dy*dy-dz*dz) +  2*mxyz*dx*dy - 2*(mxxx+mxyy)*dx*dz  - 2*(mxxy+myyy)*dy*dz

                    #ox += 21*mrrr*dx*idr9 - 9*mrrx*idr7
                    #oy += 21*mrrr*dy*idr9 - 9*mrry*idr7
                    #oz += 21*mrrr*dz*idr9 - 9*mrrz*idr7
                else:
                    ''' self contribution from the image point'''
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/dz
                    idr5 = idr*idr*idr*idr*idr

                    #mrrr = mxxz +  myyz
                    #mrrx = mxxx*(-dz*dz) + mxyy*(-dz*dz)
                    #mrry = mxxy*(-dz*dz) + myyy*(-dz*dz)
                    #mrrz = mxxz*(-dz*dz) + myyz*(-dz*dz)

                    #ox #+= - 9*mrrx*idr7
                    #oy #+= - 9*mrry*idr7
                    oz +=  12*(mxxz+myyz)*idr5

            #o[i  ]  += ox
            #o[i+Np] += oy
            o[i+xx] += oz
        return

    ## Noise
    cpdef noiseTT(self, double [:] v, double [:] r):
        """
        Compute translation Brownian motion 
        ...

        Parameters
        ----------
        v: np.array
            An array of velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        """

        cdef int i, j, Np=self.Np, xx=2*Np
        cdef double dx, dy, dz, idr, h2, hsq, idr2, idr3, idr4, idr5
        cdef double mu=self.mu, muv=2*mu*self.a*0.75, a2=self.a*self.a/3.0
        cdef double vx, vy, vz, mm=1/(.75*self.a)

        cdef double [:, :] M = self.Mobility
        cdef double [:]    Fr = np.random.normal(size=3*Np)


        for i in prange(Np, nogil=True):
            for j in range(Np):
                dx = r[i]    - r[j]
                dy = r[i+Np] - r[j+Np]
                h2=2*r[j+xx]; hsq=r[j+xx]*r[j+xx]
                if i!=j:
                    dz = r[i+xx] - r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr2=idr*idr;  idr3=idr*idr*idr
                    dx = dx*idr; dy=dy*idr; dz=dz*idr
                    #
                    M[i,    j   ] = (1 + dx*dx)*idr + a2*(2 - 6*dx*dx)*idr3
                    M[i+Np, j+Np] = (1 + dy*dy)*idr + a2*(2 - 6*dy*dy)*idr3
                    M[i+xx, j+xx] = (1 + dz*dz)*idr + a2*(2 - 6*dz*dz)*idr3
                    M[i,    j+Np] = (    dx*dy)*idr + a2*(  - 6*dx*dy)*idr3
                    M[i,    j+xx] = (    dx*dz)*idr + a2*(  - 6*dx*dz)*idr3
                    M[i+Np, j+xx] = (    dy*dz)*idr + a2*(  - 6*dy*dz)*idr3

                    ###contributions from the image
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    dx = dx*idr; dy=dy*idr; dz=dz*idr
                    idr2=idr*idr;  idr3=idr2*idr;  idr4=idr3*idr; idr5=idr4*idr

                    M[i,    j   ] += -(1 + dx*dx)*idr - a2*(2 - 6*dx*dx)*idr3
                    M[i+Np, j+Np] += -(1 + dy*dy)*idr - a2*(2 - 6*dy*dy)*idr3
                    M[i+xx, j+xx] += -(1 + dz*dz)*idr - a2*(2 - 6*dz*dz)*idr3
                    M[i,    j+Np] += -(    dx*dy)*idr - a2*(  - 6*dx*dy)*idr3
                    M[i,    j+xx] += -(    dx*dz)*idr - a2*(  - 6*dx*dz)*idr3
                    M[i+Np, j+xx] += -(    dy*dz)*idr - a2*(  - 6*dy*dz)*idr3

                    M[i,    j   ] += -h2*(dz*(1 - 3*dx*dx))*idr2
                    M[i+Np, j+Np] += -h2*(dz*(1 - 3*dy*dy))*idr2
                    M[i+xx, j+xx] += +h2*(dz*(1 - 3*dz*dz))*idr2
                    M[i,    j+Np] += -h2*(dz*(  - 3*dx*dy))*idr2
                    M[i,    j+xx] += +h2*(dz*(  - 3*dx*dz) + dx)*idr2
                    M[i+Np, j+xx] += +h2*(dz*(  - 3*dy*dz) + dy)*idr2

                    M[i,    j   ] += hsq*(2 - 6*dx*dx)*idr3
                    M[i+Np, j+Np] += hsq*(2 - 6*dy*dy)*idr3
                    M[i+xx, j+xx] +=-hsq*(2 - 6*dz*dz)*idr3
                    M[i,    j+Np] += hsq*(  - 6*dx*dy)*idr3
                    M[i,    j+xx] +=-hsq*(  - 6*dx*dz)*idr3
                    M[i+Np, j+xx] +=-hsq*(  - 6*dy*dz)*idr3

                    M[i,    j   ] += 12*a2*dz*(dz - 5*dx*dx)*idr3
                    M[i+Np, j+Np] += 12*a2*dz*(dz - 5*dy*dy)*idr3
                    M[i+xx, j+xx] +=-12*a2*dz*(dz - 5*dz*dz + 2*dz)*idr3
                    M[i,    j+Np] += 12*a2*dz*(   - 5*dx*dy)*idr3
                    M[i,    j+xx] +=-12*a2*dz*(   - 5*dx*dz + 2*dx)*idr3
                    M[i+Np, j+xx] +=-12*a2*dz*(   - 5*dy*dz + 2*dy)*idr3

                    M[i,    j   ] += h2*6*a2*(-dz   + 5*dx*dx*dz)*idr4
                    M[i+Np, j+Np] += h2*6*a2*(-dz   + 5*dy*dy*dz)*idr4
                    M[i+xx, j+xx] +=-h2*6*a2*(-3*dz + 5*dz*dz*dz)*idr4
                    M[i,    j+Np] += h2*6*a2*(        5*dx*dy*dz)*idr4
                    M[i,    j+xx] +=-h2*6*a2*(-dx   + 5*dx*dz*dz)*idr4
                    M[i+Np, j+xx] +=-h2*6*a2*(-dy   + 5*dy*dz*dz)*idr4
                else:
                    # one-body mobility
                    M[i,    j   ] = mm
                    M[i+Np, j+Np] = mm
                    M[i+xx, j+xx] = mm
                    M[i,    j+Np] = 0
                    M[i,    j+xx] = 0
                    M[i+Np, j+xx] = 0

                    ##self contribtion from the image point
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    dx = dx*idr; dy=dy*idr; dz=dz*idr
                    idr2=idr*idr;  idr3=idr2*idr;  idr4=idr3*idr; idr5=idr4*idr

                    M[i,    j   ] += -(1 + dx*dx)*idr - a2*(2 - 6*dx*dx)*idr3
                    M[i+Np, j+Np] += -(1 + dy*dy)*idr - a2*(2 - 6*dy*dy)*idr3
                    M[i+xx, j+xx] += -(1 + dz*dz)*idr - a2*(2 - 6*dz*dz)*idr3
                    M[i,    j+Np] += -(    dx*dy)*idr - a2*(  - 6*dx*dy)*idr3
                    M[i,    j+xx] += -(    dx*dz)*idr - a2*(  - 6*dx*dz)*idr3
                    M[i+Np, j+xx] += -(    dy*dz)*idr - a2*(  - 6*dy*dz)*idr3

                    M[i,    j   ] += -h2*(dz*(1 - 3*dx*dx))*idr2
                    M[i+Np, j+Np] += -h2*(dz*(1 - 3*dy*dy))*idr2
                    M[i+xx, j+xx] += +h2*(dz*(1 - 3*dz*dz))*idr2
                    M[i,    j+Np] += -h2*(dz*(  - 3*dx*dy))*idr2
                    M[i,    j+xx] += +h2*(dz*(  - 3*dx*dz) + dx)*idr2
                    M[i+Np, j+xx] += +h2*(dz*(  - 3*dy*dz) + dy)*idr2

                    M[i,    j   ] += hsq*(2 - 6*dx*dx)*idr3
                    M[i+Np, j+Np] += hsq*(2 - 6*dy*dy)*idr3
                    M[i+xx, j+xx] +=-hsq*(2 - 6*dz*dz)*idr3
                    M[i,    j+Np] += hsq*(  - 6*dx*dy)*idr3
                    M[i,    j+xx] +=-hsq*(  - 6*dx*dz)*idr3
                    M[i+Np, j+xx] +=-hsq*(  - 6*dy*dz)*idr3

                    M[i,    j   ] += 12*a2*dz*(dz - 5*dx*dx)*idr3
                    M[i+Np, j+Np] += 12*a2*dz*(dz - 5*dy*dy)*idr3
                    M[i+xx, j+xx] +=-12*a2*dz*(dz - 5*dz*dz + 2*dz)*idr3
                    M[i,    j+Np] += 12*a2*dz*(   - 5*dx*dy)*idr3
                    M[i,    j+xx] +=-12*a2*dz*(   - 5*dx*dz + 2*dx)*idr3
                    M[i+Np, j+xx] +=-12*a2*dz*(   - 5*dy*dz + 2*dy)*idr3

                    M[i,    j   ] += h2*6*a2*(-dz   + 5*dx*dx*dz)*idr4
                    M[i+Np, j+Np] += h2*6*a2*(-dz   + 5*dy*dy*dz)*idr4
                    M[i+xx, j+xx] +=-h2*6*a2*(-3*dz + 5*dz*dz*dz)*idr4
                    M[i,    j+Np] += h2*6*a2*(        5*dx*dy*dz)*idr4
                    M[i,    j+xx] +=-h2*6*a2*(-dx   + 5*dx*dz*dz)*idr4
                    M[i+Np, j+xx] +=-h2*6*a2*(-dy   + 5*dy*dz*dz)*idr4

        for i in prange(Np, nogil=True):
            for j in range(Np):
                M[i,    j   ] = muv*M[i,    j   ]
                M[i+Np, j+Np] = muv*M[i+Np, j+Np]
                M[i+xx, j+xx] = muv*M[i+xx, j+xx]
                M[i,    j+Np] = muv*M[i,    j+Np]
                M[i,    j+xx] = muv*M[i,    j+xx]
                M[i+Np, j+xx] = muv*M[i+Np, j+xx]

                M[i+Np, j   ] =     M[i,    j+Np]
                M[i+xx, j   ] =     M[i,    j+xx]
                M[i+xx, j+Np] =     M[i+Np, j+xx]

        cdef double [:, :] L = np.linalg.cholesky(self.Mobility)

        for i in prange(Np, nogil=True):
            vx=0; vy=0; vz=0;
            for j in range(Np):
                vx += L[i   , j]*Fr[j] + L[i   , j+Np]*Fr[j+Np] + L[i   , j+xx]*Fr[j+xx]
                vy += L[i+Np, j]*Fr[j] + L[i+Np, j+Np]*Fr[j+Np] + L[i+Np, j+xx]*Fr[j+xx]
                vz += L[i+xx, j]*Fr[j] + L[i+xx, j+Np]*Fr[j+Np] + L[i+xx, j+xx]*Fr[j+xx]
            v[i  ]  += vx
            v[i+Np] += vy
            v[i+xx] += vz

        return


    cpdef noiseRR(self, double [:] o, double [:] r):
        """
        Compute rotational Brownian motion 
        ...

        Parameters
        ----------
        o: np.array
            An array of angular velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        """

        cdef int i, j, Np=self.Np, xx=2*Np
        cdef double dx, dy, dz, idr, h2, hsq, idr2, idr3, idr4, idr5
        cdef double mur=self.mur, muv=0.25*sqrt(2.0)*mur, mm=4/(self.a**3)
        cdef double ox, oy, oz

        cdef double [:, :] M = self.Mobility
        cdef double [:]   Tr = np.random.normal(size=3*Np)


        for i in prange(Np, nogil=True):
            for j in range(Np):
                dx = r[i]    - r[j]
                dy = r[i+Np] - r[j+Np]
                h2=2*r[j+xx]; hsq=r[j+xx]*r[j+xx]
                if i!=j:
                    dz = r[i+xx] - r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr2=idr*idr;  idr3=idr*idr*idr
                    dx = dx*idr; dy=dy*idr; dz=dz*idr
                    #
                    M[i,    j   ] = (2 - 6*dx*dx)*idr3
                    M[i+Np, j+Np] = (2 - 6*dy*dy)*idr3
                    M[i+xx, j+xx] = (2 - 6*dz*dz)*idr3
                    M[i,    j+Np] = (  - 6*dx*dy)*idr3
                    M[i,    j+xx] = (  - 6*dx*dz)*idr3
                    M[i+Np, j+xx] = (  - 6*dy*dz)*idr3

                    ###contributions from the image
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    dx = dx*idr; dy=dy*idr; dz=dz*idr
                    idr2=idr*idr;  idr3=idr2*idr;  idr4=idr3*idr

                    M[i,    j   ] += -(2 - 6*dx*dx)*idr3
                    M[i+Np, j+Np] += -(2 - 6*dy*dy)*idr3
                    M[i+xx, j+xx] += -(2 - 6*dz*dz)*idr3
                    M[i,    j+Np] += -(  - 6*dx*dy)*idr3
                    M[i,    j+xx] += -(  - 6*dx*dz)*idr3
                    M[i+Np, j+xx] += -(  - 6*dy*dz)*idr3

                    M[i,    j   ] += 12*dz*(dz - 5*dx*dx       )*idr3
                    M[i+Np, j+Np] += 12*dz*(dz - 5*dy*dy       )*idr3
                    M[i+xx, j+xx] +=-12*dz*(dz - 5*dz*dz + 2*dz)*idr3
                    M[i,    j+Np] += 12*dz*(   - 5*dx*dy       )*idr3
                    M[i,    j+xx] +=-12*dz*(   - 5*dx*dz + 2*dx)*idr3
                    M[i+Np, j+xx] +=-12*dz*(   - 5*dy*dz + 2*dy)*idr3

                    M[i,    j   ] += h2*6*(-dz   + 5*dx*dx*dz)*idr4
                    M[i+Np, j+Np] += h2*6*(-dz   + 5*dy*dy*dz)*idr4
                    M[i+xx, j+xx] +=-h2*6*(-3*dz + 5*dz*dz*dz)*idr4
                    M[i,    j+Np] += h2*6*(        5*dx*dy*dz)*idr4
                    M[i,    j+xx] +=-h2*6*(-dx   + 5*dx*dz*dz)*idr4
                    M[i+Np, j+xx] +=-h2*6*(-dy   + 5*dy*dz*dz)*idr4
                else:
                    # one-body mobility
                    M[i,    j   ] = mm
                    M[i+Np, j+Np] = mm
                    M[i+xx, j+xx] = mm
                    M[i,    j+Np] = 0
                    M[i,    j+xx] = 0
                    M[i+Np, j+xx] = 0

                    ##self contribtion from the image point
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    dx = dx*idr; dy=dy*idr; dz=dz*idr
                    idr2=idr*idr;  idr3=idr2*idr;  idr4=idr3*idr

                    M[i,    j   ] += -(2 - 6*dx*dx)*idr3
                    M[i+Np, j+Np] += -(2 - 6*dy*dy)*idr3
                    M[i+xx, j+xx] += -(2 - 6*dz*dz)*idr3
                    M[i,    j+Np] += -(  - 6*dx*dy)*idr3
                    M[i,    j+xx] += -(  - 6*dx*dz)*idr3
                    M[i+Np, j+xx] += -(  - 6*dy*dz)*idr3

                    M[i,    j   ] += 12*dz*(dz - 5*dx*dx       )*idr3
                    M[i+Np, j+Np] += 12*dz*(dz - 5*dy*dy       )*idr3
                    M[i+xx, j+xx] +=-12*dz*(dz - 5*dz*dz + 2*dz)*idr3
                    M[i,    j+Np] += 12*dz*(   - 5*dx*dy       )*idr3
                    M[i,    j+xx] +=-12*dz*(   - 5*dx*dz + 2*dx)*idr3
                    M[i+Np, j+xx] +=-12*dz*(   - 5*dy*dz + 2*dy)*idr3

                    M[i,    j   ] += h2*6*(-dz   + 5*dx*dx*dz)*idr4
                    M[i+Np, j+Np] += h2*6*(-dz   + 5*dy*dy*dz)*idr4
                    M[i+xx, j+xx] +=-h2*6*(-3*dz + 5*dz*dz*dz)*idr4
                    M[i,    j+Np] += h2*6*(        5*dx*dy*dz)*idr4
                    M[i,    j+xx] +=-h2*6*(-dx   + 5*dx*dz*dz)*idr4
                    M[i+Np, j+xx] +=-h2*6*(-dy   + 5*dy*dz*dz)*idr4

        for i in prange(Np, nogil=True):
            for j in range(Np):
                M[i,    j   ] = muv*M[i,    j   ]
                M[i+Np, j+Np] = muv*M[i+Np, j+Np]
                M[i+xx, j+xx] = muv*M[i+xx, j+xx]
                M[i,    j+Np] = muv*M[i,    j+Np]
                M[i,    j+xx] = muv*M[i,    j+xx]
                M[i+Np, j+xx] = muv*M[i+Np, j+xx]

                M[i+Np, j   ] =     M[i,    j+Np]
                M[i+xx, j   ] =     M[i,    j+xx]
                M[i+xx, j+Np] =     M[i+Np, j+xx]

        cdef double [:, :] L = muv*np.linalg.cholesky(self.Mobility)
        for i in prange(Np, nogil=True):
            ox=0; oy=0; oz=0;
            for j in range(Np):
                ox += L[i   , j]*Tr[j] + L[i   , j+Np]*Tr[j+Np] + L[i   , j+xx]*Tr[j+xx]
                oy += L[i+Np, j]*Tr[j] + L[i+Np, j+Np]*Tr[j+Np] + L[i+Np, j+xx]*Tr[j+xx]
                oz += L[i+xx, j]*Tr[j] + L[i+xx, j+Np]*Tr[j+Np] + L[i+xx, j+xx]*Tr[j+xx]
            o[i  ]  += ox
            o[i+Np] += oy
            o[i+xx] += oz
        return




@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
@cython.wraparound(False)
cdef class Flow:
    """
    Flow at given points
    
    ...

    Parameters
    ----------
    radius: float
        Radius of the particles.    
    particles: int
        Number of particles 
    viscosity: viscosity of the fluid 
    gridpoints: int 
        Number of grid points

    Examples
    --------
    An example of the RBM

    """

    def __init__(self, radius=1, particles=1, viscosity=1, gridpoints=32):
        self.a  = radius
        self.Np = particles
        self.Nt = gridpoints
        self.eta= viscosity



    cpdef flowField1s(self, double [:] vv, double [:] rt, double [:] r, double [:] F):
        """
        Compute flow field at field points due body forces
        ...

        Parameters
        ----------
        vv: np.array
            An array of flow at field points
            An array of size 3*Nt,
        rt: np.array
            An array of field points
            An array of size 3*Nt,
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of body force
            An array of size 3*Np,
    
        Examples
        --------
        An example of the Flow field due to $1s$ mode of force per unit area

        >>> import pystokes, numpy as np, matplotlib.pyplot as plt
        >>> 
        >>> # particle radius, self-propulsion speed, number and fluid viscosity
        >>> b, eta, Np = 1.0, 1.0/6.0, 1
        >>> 
        >>> # initialize
        >>> r, p = np.array([0.0, 0.0, 3.4]), np.array([0.0, 1.0, 0])
        >>> F1s  = pystokes.utils.irreducibleTensors(1, p)
        >>> 
        >>> # space dimension , extent , discretization
        >>> dim, L, Ng = 3, 10, 64;
        >>> 
        >>> # instantiate the Flow class
        >>> flow = pystokes.wallBounded.Flow(radius=b, particles=Np, viscosity=eta, gridpoints=Ng*Ng)
        >>> 
        >>> # create grid, evaluate flow and plot
        >>> rr, vv = pystokes.utils.gridYZ(dim, L, Ng)
        >>> flow.flowField1s(vv, rr, r, F1s)
        >>> pystokes.utils.plotStreamlinesYZsurf(vv, rr, r, offset=6-1, density=1.4, title='1s')
        """

        cdef int i, j, Np=self.Np, Nt=self.Nt, xx=2*Np
        cdef double dx, dy, dz, idr, idr3, idr5, Fdotidr, tempF, hsq, h2
        cdef double vx, vy, vz, muv=1.0/(8*PI*self.eta), a2=self.a*self.a/6.0

        for i in prange(Nt, nogil=True):
            vx=0; vy=0; vz=0;
            for j in range(Np):
                h2  =  2*r[j+xx]; hsq=r[j+xx]*r[j+xx]
                dx = rt[i]    - r[j]
                dy = rt[i+Nt] - r[j+Np]
                dz = rt[i+2*Nt]  - r[j+xx]
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr3=idr*idr*idr
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr3=idr*idr*idr
                Fdotidr = (F[j] * dx + F[j+Np] * dy + F[j+xx] * dz)*idr*idr
                #
                vx += (F[j]   +Fdotidr*dx)*idr + a2*(2*F[j]   -6*Fdotidr*dx)*idr3
                vy += (F[j+Np]+Fdotidr*dy)*idr + a2*(2*F[j+Np]-6*Fdotidr*dy)*idr3
                vz += (F[j+xx]+Fdotidr*dz)*idr + a2*(2*F[j+xx]-6*Fdotidr*dz)*idr3

                ##contributions from the image
                dz = rt[i+2*Nt] + r[j+xx]
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr3 = idr*idr*idr
                idr5 = idr3*idr*idr
                Fdotidr = ( F[j]*dx + F[j+Np]*dy + F[j+xx]*dz )*idr*idr

                vx += -(F[j]   +Fdotidr*dx)*idr - a2*(2*F[j]   -6*Fdotidr*dx)*idr3
                vy += -(F[j+Np]+Fdotidr*dy)*idr - a2*(2*F[j+Np]-6*Fdotidr*dy)*idr3
                vz += -(F[j+xx]+Fdotidr*dz)*idr - a2*(2*F[j+xx]-6*Fdotidr*dz)*idr3

                tempF  = -F[j+xx]     # F_i = M_ij F_j, reflection of the strength
                Fdotidr = ( F[j]*dx + F[j+Np]*dy + tempF*dz )*idr*idr

                vx += -h2*(dz*(F[j]   - 3*Fdotidr*dx) + tempF*dx)*idr3
                vy += -h2*(dz*(F[j+Np]- 3*Fdotidr*dy) + tempF*dy)*idr3
                vz += -h2*(dz*(tempF  - 3*Fdotidr*dz) + tempF*dz)*idr3 + h2*Fdotidr*idr

                vx += hsq*( 2*F[j]   - 6*Fdotidr*dx )*idr3
                vy += hsq*( 2*F[j+Np]- 6*Fdotidr*dy )*idr3
                vz += hsq*( 2*tempF  - 6*Fdotidr*dz )*idr3

                vx += 12*a2*dz*( dz*F[j]   - 5*dz*Fdotidr*dx + 2*tempF*dx )*idr5
                vy += 12*a2*dz*( dz*F[j+Np]- 5*dz*Fdotidr*dy + 2*tempF*dy )*idr5
                vz += 12*a2*dz*( dz*tempF  - 5*dz*Fdotidr*dz + 2*tempF*dz )*idr5

                vx += -h2*6*a2*(dz*F[j]   -5*Fdotidr*dx*dz + tempF*dx)*idr5
                vy += -h2*6*a2*(dz*F[j+Np]-5*Fdotidr*dy*dz+ tempF*dy)*idr5
                vz += -h2*6*a2*(dz*tempF  -5*Fdotidr*dz*dz + tempF*dz)*idr5 -6*a2*h2*Fdotidr*idr3

            vv[i  ]    += muv*vx
            vv[i+Nt]   += muv*vy
            vv[i+2*Nt] += muv*vz
        return 
    
    
    cpdef flowField2a(  self, double [:] vv, double [:] rt, double [:] r, double [:] T):
        """
        Compute flow field at field points due body Torque 
        ...

        Parameters
        ----------
        vv: np.array
            An array of flow at field points
            An array of size 3*Nt,
        rt: np.array
            An array of field points
            An array of size 3*Nt,
        r: np.array
            An array of positions
            An array of size 3*Np,
        T: np.array
            An array of body torque
            An array of size 3*Np,
    
        Examples
        --------
        An example of the RBM 

        # Example 1: Flow field due to $2a$ mode of force per unit area
        >>> import pystokes, numpy as np, matplotlib.pyplot as plt
        >>> 
        >>> # particle radius, self-propulsion speed, number and fluid viscosity
        >>> b, eta, Np = 1.0, 1.0/6.0, 1
        >>> 
        >>> # initialize
        >>> r, p = np.array([0.0, 0.0, 3.4]), np.array([0.0, 1.0, 0])
        >>> V2a  = pystokes.utils.irreducibleTensors(1, p)
        >>> 
        >>> # space dimension , extent , discretization
        >>> dim, L, Ng = 3, 10, 64;
        >>> 
        >>> # instantiate the Flow class
        >>> flow = pystokes.wallBounded.Flow(radius=b, particles=Np, viscosity=eta, gridpoints=Ng*Ng)
        >>> 
        >>> # create grid, evaluate flow and plot
        >>> rr, vv = pystokes.utils.gridYZ(dim, L, Ng)
        >>> flow.flowField2a(vv, rr, r, V2a)
        >>> pystokes.utils.plotStreamlinesYZsurf(vv, rr, r, offset=6-1, density=1.4, title='2s')
        """

        cdef int Np = self.Np, i, j, xx=2*Np, Nt=self.Nt
        cdef double dx, dy, dz, idr, idr3, rlz, Tdotidr, h2, 
        cdef double vx, vy, vz, muv = 1.0/(8*PI*self.eta)
#        cdef int i, j, Np=self.Np, Nt=self.Nt, xx=2*Np
#        cdef double dx, dy, dz, idr, idr3, idr5, Fdotidr, tempF, hsq, h2
#        cdef double vx, vy, vz, muv=1.0/(8*PI*self.eta), a2=self.a*self.a/6.0
 
        for i in prange(Nt, nogil=True):
            vx=0; vy=0; vz=0;
            for j in range(Np):
                dx = rt[i]   - r[j]   
                dy = rt[i+Nt] - r[j+Np]   
                h2 = 2*rt[i+Nt*2]
                #contributions from the source 
                dz = rt[i+2*Nt] - r[j+xx] 
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr3 = idr*idr*idr
                 
                vx += (T[j+Np]*dz - T[j+xx]*dy )*idr3
                vy += (T[j+xx]*dx - T[j]   *dz )*idr3
                vz += (T[j]   *dy - T[j+Np]*dx )*idr3
                    
                #contributions from the image 
                dz = rt[i+2*Nt] + r[j+xx]            
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr3 = idr*idr*idr
                
                vx += -(T[j+Np]*dz - T[j+xx]*dy )*idr3
                vy += -(T[j+xx]*dx - T[j]   *dz )*idr3
                vz += -(T[j]   *dy - T[j+Np]*dx )*idr3
                
                rlz = (dx*T[j+Np] - dy*T[j])*idr*idr
                vx += (h2*(T[j+Np]-3*rlz*dx) + 6*dz*dx*rlz)*idr3
                vy += (h2*(-T[j]  -3*rlz*dy) + 6*dz*dy*rlz)*idr3
                vz += (h2*(       -3*rlz*dz) + 6*dz*dz*rlz)*idr3

                dz = rt[i+2*Nt] + r[j+xx]            
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr3 = idr*idr*idr
                
                vx += -(T[j+Np]*dz )*idr3
                vy += -(- T[j] *dz )*idr3
                
                vx += h2*T[j+Np]*idr3
                vy += -h2*T[j]*idr3

            vv[i  ]  += muv*vx 
            vv[i+Nt] += muv*vy
            vv[i+2*Nt] += muv*vz
        return 
    
   
    cpdef flowField2s(self, double [:] vv, double [:] rt, double [:] r, double [:] S):
        """
        Compute flow field at field points  due to 2s mode of the slip 
        ...

        Parameters
        ----------
        vv: np.array
            An array of flow at field points
            An array of size 3*Nt,
        rt: np.array
            An array of field points
            An array of size 3*Nt,
        r: np.array
            An array of positions
            An array of size 3*Np,
        S: np.array
            An array of 2s mode of the slip
            An array of size 5*Np,
        
        Examples
        --------
        An example of the Flow field due to $3t$ mode of active slip

        >>> import pystokes, numpy as np, matplotlib.pyplot as plt
        >>> 
        >>> # particle radius, self-propulsion speed, number and fluid viscosity
        >>> b, eta, Np = 1.0, 1.0/6.0, 1
        >>> 
        >>> # initialize
        >>> r, p = np.array([0.0, 0.0, 3.4]), np.array([0.0, 1.0, 0])
        >>> V3t  = pystokes.utils.irreducibleTensors(1, p)
        >>> 
        >>> # space dimension , extent , discretization
        >>> dim, L, Ng = 3, 10, 64;
        >>> 
        >>> # instantiate the Flow class
        >>> flow = pystokes.wallBounded.Flow(radius=b, particles=Np, viscosity=eta, gridpoints=Ng*Ng)
        >>> 
        >>> # create grid, evaluate flow and plot
        >>> rr, vv = pystokes.utils.gridYZ(dim, L, Ng)
        >>> flow.flowField3t(vv, rr, r, V3t)
        >>> pystokes.utils.plotStreamlinesYZsurf(vv, rr, r, offset=6-1, density=1.4, title='1s')
        """
        cdef int Np=self.Np,  Nt=self.Nt, xx=2*Np, xx1=3*Np, xx2=4*Np
        cdef int i, j  
        cdef double dx, dy, dz, idr, idr2, idr3, idr5, idr7, aidr2, trS, h2, hsq
        cdef double sxx, syy, szz, sxy, syx, syz, szy, sxz, szx, srr, srx, sry, srz
        cdef double Sljrlx, Sljrly, Sljrlz, Sljrjx, Sljrjy, Sljrjz 
        cdef double vx, vy, vz, mus = (28.0*self.a**3)/24 

        for i in prange(Nt, nogil=True):
            vx=0; vy=0; vz=0;
            for j in  range(Np):
                sxx = S[j]  ; syy = S[j+Np]; szz = -sxx-syy;
                sxy = S[j+xx]; syx = sxy;
                sxz = S[j+xx1]; szx = sxz;
                syz = S[j+xx2]; szy = syz;
                
                dx = rt[i]   - r[j]
                dy = rt[i+Nt] - r[j+Np]
                dz = rt[i+2*Nt] - r[j+xx] 
                h2 = 2*r[j+xx]; hsq = r[j+xx]*r[j+xx];
                idr  = 1.0/sqrt( dx*dx + dy*dy + dz*dz );
                idr2 = idr*idr; idr3 = idr2*idr; idr5 = idr3*idr2; idr7 = idr5*idr2;
                srx = (sxx*dx +  sxy*dy + sxz*dz ); 
                sry = (sxy*dx +  syy*dy + syz*dz );
                srz = (sxz*dx +  syz*dy + szz*dz );
                srr = sxx*dx*dx + syy*dy*dy + szz*dz*dz + 2*sxy*dx*dy + 2*sxz*dx*dz + 2*syz*dy*dz;
                
                ## contributions from the source 
                vx += 3*srr*dx*idr5;
                vy += 3*srr*dy*idr5;
                vz += 3*srr*dz*idr5;
                 
                ## contributions from the image 
                dz = rt[i+2*Nt]+r[j+xx]
                idr  = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr2 = idr*idr; idr3 = idr2*idr; idr5 = idr3*idr2; idr7 = idr5*idr2;
                
                #reflecting the first index of stresslet, S_jl M_lm
                sxz=-sxz; syz=-syz; szz=-szz;     trS=sxx+syy+szz; 
                Sljrlx = sxx*dx +  sxy*dx + sxz*dx ; 
                Sljrly = syx*dy +  syy*dy + syz*dy ;
                Sljrlz = szx*dz +  szy*dz + szz*dz ;
                Sljrjx = sxx*dx +  sxy*dy + sxz*dz ; 
                Sljrjy = syx*dx +  syy*dy + syz*dz ;
                Sljrjz = szx*dx +  szy*dy + szz*dz ;
                srr = (sxx*dx*dx + syy*dy*dy + szz*dz*dz +  2*sxy*dx*dy)*idr2 ;
                srx = sxx*dx + syx*dy+szx*dz;
                sry = sxy*dx + syy*dy+szy*dz;
                srz = sxz*dx + syz*dy+szz*dz;
                
                vx += (-Sljrlx + Sljrjx + trS*dx -3*srr*dx)*idr3 ;
                vy += (-Sljrly + Sljrjy + trS*dy -3*srr*dy)*idr3 ;
                vz += (-Sljrlz + Sljrjz + trS*dz -3*srr*dz)*idr3 ;
                
                vx += -2*(dz*(sxz-3*srz*dx*idr2)+ szz*dx)*idr3;
                vy += -2*(dz*(syz-3*srz*dy*idr2)+ szz*dy)*idr3;
                vz += -2*(dz*(szz-3*srz*dz*idr2)+ szz*dz - srz)*idr3;
                
                vx += h2*( sxz-3*srz*dx*idr2)*idr3;
                vy += h2*( syz-3*srz*dy*idr2)*idr3;
                vz += h2*( szz-3*srz*dz*idr2)*idr3;
                
                #reflecting both the indices of stresslet, S_jl M_lm M_jk
                szx = -szx ; syz = -syz; szz = -szz;
                srx = sxx*dx + syx*dy+szx*dz;
                sry = sxy*dx + syy*dy+szy*dz;
                srz = sxz*dx + syz*dy+szz*dz;
                srr = (sxx*dx*dx + syy*dy*dy + szz*dz*dz + 2*sxy*dx*dy + 2*sxz*dx*dz + 2*syz*dy*dz)*idr2;
                
                vx += h2*(dz*(-6*srx + 15*srr*dx)-3*srz*dx)*idr5 + h2*(sxz)*idr3 ;
                vy += h2*(dz*(-6*sry + 15*srr*dy)-3*srz*dy)*idr5 + h2*(syz)*idr3 ;
                vz += h2*(dz*(-6*srz + 15*srr*dz)-3*srz*dz)*idr5 + h2*(szz + 3*srr)*idr3;

                vx += hsq*(12*srx - 30*srr*dx)*idr5 ;
                vy += hsq*(12*sry - 30*srr*dy)*idr5 ;
                vz += hsq*(12*srz - 30*srr*dz)*idr5 ;
     
            vv[i  ]    += mus*vx
            vv[i+Nt]   += mus*vy
            vv[i+2*Nt] += mus*vz
        return

   
    cpdef flowField3t(self, double [:] vv, double [:] rt, double [:] r, double [:] D):
        """
        Compute flow field at field points due to 3t mode of the slip 
        ...

        Parameters
        ----------
        vv: np.array
            An array of flow at field points
            An array of size 3*Nt,
        rt: np.array
            An array of field points
            An array of size 3*Nt,
        r: np.array
            An array of positions
            An array of size 3*Np,
        D: np.array
            An array of 3t mode of the slip
            An array of size 3*Np,
 
        Examples
        --------
        An example of the Flow field due to $3t$ mode of active slip

        >>> import pystokes, numpy as np, matplotlib.pyplot as plt
        >>> 
        >>> # particle radius, self-propulsion speed, number and fluid viscosity
        >>> b, eta, Np = 1.0, 1.0/6.0, 1
        >>> 
        >>> # initialize
        >>> r, p = np.array([0.0, 0.0, 3.4]), np.array([0.0, 1.0, 0])
        >>> V3t  = pystokes.utils.irreducibleTensors(1, p)
        >>> 
        >>> # space dimension , extent , discretization
        >>> dim, L, Ng = 3, 10, 64;
        >>> 
        >>> # instantiate the Flow class
        >>> flow = pystokes.wallBounded.Flow(radius=b, particles=Np, viscosity=eta, gridpoints=Ng*Ng)
        >>> 
        >>> # create grid, evaluate flow and plot
        >>> rr, vv = pystokes.utils.gridYZ(dim, L, Ng)
        >>> flow.flowField3t(vv, rr, r, V3t)
        >>> pystokes.utils.plotStreamlinesYZsurf(vv, rr, r, offset=6-1, density=1.4, title='1s')
        """

        cdef int i, j, Np=self.Np, Nt=self.Nt, xx=2*Np
        cdef double dx, dy, dz, idr, idr3, idr5, Ddotidr, tempD, hsq, h2
        cdef double vx, vy, vz, mud = 3.0*self.a*self.a*self.a/5, muv = -1.0*(self.a**5)/10

        for i in prange(Nt, nogil=True):
            vx=0; vy=0; vz=0;
            for j in range(Np):
                h2 = 2*r[j+xx]; hsq = r[j+xx]*r[j+xx];
                dx = rt[i]      - r[j]
                dy = rt[i+Nt]   - r[j+Np]
                dz = rt[i+2*Nt] - r[j+xx] 
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr3=idr*idr*idr
                Ddotidr = (D[j]*dx + D[j+Np]*dy + D[j+xx]*dz)*idr*idr
                #
                vx += (2*D[j]    - 6*Ddotidr*dx)*idr3
                vy += (2*D[j+Np] - 6*Ddotidr*dy)*idr3
                vz += (2*D[j+xx] - 6*Ddotidr*dz)*idr3
                
                ##contributions from the image 
                dz = rt[i+2*Nt] + r[j+xx]        
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr3 = idr*idr*idr
                idr5 = idr3*idr*idr
                Ddotidr = (D[j]*dx + D[j+Np]*dy + D[j+xx]*dz)*idr*idr
                
                vx += -(2*D[j]    - 6*Ddotidr*dx )*idr3
                vy += -(2*D[j+Np] - 6*Ddotidr*dy )*idr3
                vz += -(2*D[j+xx] - 6*Ddotidr*dz )*idr3
                
                tempD = -D[j+xx]     # D_i = M_ij D_j, reflection of the strength
                Ddotidr = ( D[j]*dx + D[j+Np]*dy + tempD*dz )*idr*idr
                
                vx += 12*dz*( dz*D[j]   - 5*dz*Ddotidr*dx + 2*tempD*dx )*idr5
                vy += 12*dz*( dz*D[j+Np]- 5*dz*Ddotidr*dy + 2*tempD*dy )*idr5
                vz += 12*dz*( dz*tempD  - 5*dz*Ddotidr*dz + 2*tempD*dz )*idr5

                vx += -6*h2*(dz*D[j]   -5*Ddotidr*dx*dz + tempD*dx)*idr5
                vy += -6*h2*(dz*D[j+Np]-5*Ddotidr*dy*dz + tempD*dy)*idr5
                vz += -6*h2*(dz*tempD  -5*Ddotidr*dz*dz + tempD*dz)*idr5 -6*h2*Ddotidr*idr3
                
            vv[i  ]    += muv*vx
            vv[i+Nt]   += muv*vy
            vv[i+2*Nt] += muv*vz
        return 
    




cimport cython
from libc.math cimport sqrt
from cython.parallel import prange
cdef double PI = 3.14159265359
import numpy as np
cimport numpy as np

@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
@cython.wraparound(False)
cdef class Phoresis:
    """
    Phoresis 
    
    ...

    Parameters
    ----------
    radius: float
        Radius of the particles.    
    particles: int
        Number of particles 
    phoreticConstant: float 
        PhoreticConstant 
    Examples
    --------
    An example of Phoresis

    """
    def __init__(self, radius=1, particles=1, phoreticConstant=1.0):
        self.a  = radius
        self.Np = particles
        self.D  = phoreticConstant

    cpdef elastance00(self, double [:] C0, double [:] r, double [:] J0):
        cdef int i, j, Np=self.Np, xx=2*Np
        cdef double dx, dy, dz, idr, h2, hsq, idr2, A1
        cdef double vx, vy, vz, ii , cc=0 ## cc is the concentration constant
        cdef double mud = J0[0]/(4*PI*self.D)

        for i in prange(Np, nogil=True):
            cc=0;
            for j in range(Np):
                if i!=j:
                    dx = r[i]    - r[j]
                    dy = r[i+Np]  - r[j+Np]
                    dz = r[i+xx]  - r[j+xx]
                    idr = 1.0/sqrt(dx*dx + dy*dy + dz*dz)
                    cc += idr

                    ###contributions from the image 
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    cc += idr
                else:
                    ''' self contribution from the image point''' 
                    dz = r[i+xx] + r[j+xx]    
                    cc += 1/(dz)
    
            C0[i  ]  += mud*cc 
        return 
    

    cpdef elastance10(self, double [:] C1, double [:] r, double [:] J0):
        cdef int i, j, Np=self.Np, xx=2*Np
        cdef double dx, dy, dz, idr, h2, hsq, idr2, A1
        cdef double vx, vy, vz, ii , cc=1 ## cc is the concentration constant
        cdef double mud = J0[0]/(4*PI*self.D)

        for i in prange(Np, nogil=True):
            vx=0; vy=0; vz=0;
            for j in range(Np):
                if i!=j:
                    dx = r[i]    - r[j]
                    dy = r[i+Np]  - r[j+Np]
                    dz = r[i+xx]  - r[j+xx]
                    idr = 1.0/sqrt(dx*dx + dy*dy + dz*dz)
                    A1 = idr*idr*idr
                    vx += A1*dx
                    vy += A1*dy
                    vz += A1*dz

                    ###contributions from the image 
                    dz = r[i+xx] + r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    A1  = idr*idr*idr
                    vx += A1*dx
                    vy += A1*dy
                    vz += A1*dz
                else:
                    ''' self contribution from the image point''' 
                    dz = r[i+xx] + r[j+xx]    
                    vz += 1/(dz*dz)
    
            C1[i  ]  += mud*vx 
            C1[i+Np] += mud*vy
            C1[i+xx] += mud*vz
        return 


    cpdef elastance11(self, double [:] C1, double [:] r, double [:] J1):
        cdef int i, j, Np=self.Np, xx=2*Np
        cdef double dx, dy, dz, idr, h2, hsq, idr2, A1, B1
        cdef double vx, vy, vz, ii , cc=1 ## cc is the concentration constant
        cdef double mud = 5/(PI*PI**self.a)
 
        for i in prange(Np, nogil=True):
            vx=0; vy=0; vz=0;
            for j in range(Np):
                if i!=j:
                    dx = r[i]    - r[j]
                    dy = r[i+Np] - r[j+Np]
                    dz = r[i+xx] - r[j+xx] 
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz)
                    A1 = idr*idr*idr
                    B1  = 3*(J1[j]*dx + J1[j+Np]*dy + J1[j+xx]*dz)*idr*idr
                    vx += A1*(J1[j]    - B1*dx)
                    vy += A1*(J1[j+Np] - B1*dy)
                    vz += A1*(J1[j+xx] - B1*dz)

                    ###contributions from the image 
                    dz = r[i+xx]  + r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz)
                    A1 = idr*idr*idr
                    B1  = 3*(J1[j]*dx + J1[j+Np]*dy + J1[j+xx]*dz)*idr*idr
                    vx += A1*(J1[j]    - B1*dx)
                    vy += A1*(J1[j+Np] - B1*dy)
                    vz += A1*(J1[j+xx] - B1*dz)

                else:
                   ''' self contribution from the image point'''
                   dz = 2*r[i+xx] ; A1=1/(dz*dz*dz)
                   vx += A1*(   J1[j]    )
                   vy += A1*(   J1[j+Np] )
                   vy += A1*(-2*J1[j+xx] )

            C1[i  ]  += mud*J1[i  ]  + mud*vx
            C1[i+Np] += mud*J1[i+Np] + mud*vy
            C1[i+xx] += mud*J1[i+xx] + mud*vz
        return




@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
@cython.wraparound(False)
cdef class Field:
    """
    Phoretic field at given points
    
    ...

    Parameters
    ----------
    radius: float
        Radius of the particles.    
    particles: int
        Number of particles 
    phoreticConstant: float
        PhoreticConstant of the fluid 
    gridpoints: int 
        Number of grid points
    """

    def __init__(self, radius=1, particles=1, phoreticConstant=1, gridpoints=32):
        self.a  = radius
        self.Np = particles
        self.Nt = gridpoints
        self.D  = phoreticConstant


    cpdef phoreticField0(self, double [:] c, double [:] rt, double [:] r, double [:] J0):
        cdef int Np = self.Np, Nt = self.Nt, xx = 2*Np
        cdef int i, j
        cdef double dx, dy, dz, idr, idr3, cc, mu1=J0[0]/(4*PI*self.D)
 
        for i in prange(Nt, nogil=True):
            cc=0
            for j in range(Np):
                dx = rt[i]      - r[j]
                dy = rt[i+Nt]   - r[j+Np]
                dz = rt[i+2*Nt] - r[j+xx] 
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                cc +=idr
                
                ##contributions from the image 
                dz  = rt[i+2*Nt] + r[j+xx]        
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                cc  += idr

            c[i] += mu1*cc
        return 
    
    
    cpdef phoreticField1(self, double [:] c, double [:] rt, double [:] r, double [:] J1):
        cdef int Np = self.Np, Nt = self.Nt, xx = 2*Np
        cdef int i, j
        cdef double dx, dy, dz, idr, idr3, cc, mu1=1.0/(4*PI*self.D)
 
        for i in prange(Nt, nogil=True):
            cc=0
            for j in range(Np):
                dx = rt[i]      - r[j]
                dy = rt[i+Nt]   - r[j+Np]
                dz = rt[i+2*Nt] - r[j+2*Np] 

                idr  = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr3 = idr*idr*idr 

                cc  += idr3*(dx*J1[j] + dy*J1[j+Np] + dz*J1[j+2*Np])
                
                ##contributions from the image 
                dz   = rt[i+2*Nt] + r[j+xx]        
                idr  = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr3 = idr*idr*idr 

                cc  += idr3*(dx*J1[j] + dy*J1[j+Np] + dz*J1[j+2*Np])

            c[i] += mu1*cc
        return 
    
    
    cpdef phoreticField2(self, double [:] c, double [:] rt, double [:] r, double [:] J2):
        cdef int Np = self.Np, Nt = self.Nt, xx = 2*Np
        cdef int i, j
        cdef double dx, dy, dz, idr, idr5, cc, mu1=1.0/(4*PI*self.D), jxx, jyy, jxy, jyz, jxz
 
        for i in prange(Nt, nogil=True):
            cc=0
            for j in range(Np):
                jxx = J2[j]
                jyy = J2[j+Np]
                jxy = J2[j+2*Np]
                jxz = J2[j+3*Np]
                jyz = J2[j+4*Np]

                dx = rt[i]      - r[j]
                dy = rt[i+Nt]   - r[j+Np]
                dz = rt[i+2*Nt] - r[j+2*Np] 

                idr  = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr5 = idr*idr*idr*idr*idr 

                cc  += jxx*dx*dx + jyy*dy*dy + (-jxx-jyy)*dz*dz
                cc  +=(jxy*dx*dy + jxz*dx*dz+ jyz*dy*dz)*2
                
                ##contributions from the image 
                dz   = rt[i+2*Nt] + r[j+xx]        
                idr  = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr  = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr5 = idr*idr*idr*idr*idr 

                cc  += jxx*dx*dx + jyy*dy*dy + (-jxx-jyy)*dz*dz
                cc  +=(jxy*dx*dy + jxz*dx*dz+ jyz*dy*dz)*2

            c[i] += mu1*cc
        return 



cimport cython
from libc.math cimport sqrt, pow
from cython.parallel import prange


@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
@cython.wraparound(False)
cdef class Forces:
    """
    Computes forces in a system of colloidal particles

    Methods in the Forces class take input,
        * arrays of positions, forces 
        * parameters for a given potential
    
    The array of forces is then update by each method. 

    ...

    Parameters
    ----------
    particles: int
        Number of particles (Np)


    """
    def __init__(self, particles=1):
        self.Np = particles


    cpdef lennardJones(self, double [:] F, double [:] r, double lje=0.01, double ljr=3):
        """
        The standard Lennard-Jones potential truncated at the minimum (aslo called WCA potential)
        
            We choose \phi(r) = lje/12 (rr^12 - 2*rr^6 ) + lje/12,  as the standard WCA potential.
            ljr: minimum of the LJ potential and rr=ljr/r.

        ...

        Parameters
        ----------
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
        lje: float
            Strength of the LJ 
        ljr: float
            Range of the LJ 


        """

        cdef int Np = self.Np, i, j, xx = 2*Np
        cdef double dx, dy, dz, dr2, idr, rminbyr, fac, fx, fy, fz

        for i in prange(Np,nogil=True):
            fx = 0.0; fy = 0.0; fz = 0.0;
            for j in range(Np):
                dx = r[i   ] - r[j   ]
                dy = r[i+Np] - r[j+Np]
                dz = r[i+xx] - r[j+xx]
                dr2 = dx*dx + dy*dy + dz*dz
                if i != j and dr2 < (ljr*ljr):
                    idr     = 1.0/sqrt(dr2)
                    rminbyr = ljr*idr
                    fac   = lje*(pow(rminbyr, 12) - pow(rminbyr, 6))*idr*idr

                    fx += fac*dx
                    fy += fac*dy
                    fz += fac*dz
            F[i]    += fx
            F[i+Np] += fy
            F[i+xx] += fz
        return


    cpdef lennardJonesWall(self, double [:] F, double [:] r, double lje=0.0100, double ljr=3, double wlje=0.01, double wljr=3.0):
        """
        The standard Lennard-Jones potential truncated at the minimum (aslo called WCA potential)
        
            We choose \phi(r) = lje/12 (rr^12 - 2*rr^6 ) + lje/12,  as the standard WCA potential.
            ljr: minimum of the LJ potential and rr=ljr/r.

        ...

        Parameters
        ----------
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
        lje: float
            Strength of the LJ 
        ljr: float
            Range of the LJ 


        """

        cdef int Np=self.Np, i, j, xx=2*Np
        cdef double dx, dy, dz, dr, idr, rminbyr, fac, fx, fy, fz, hh

        for i in prange(Np, nogil=True):
            fx = 0.0; fy = 0.0; fz = 0.0;
            hh = r[i+xx]
            if hh<wljr:
                idr = 1/hh
                rminbyr = wljr*idr
                fac   = wlje*(pow(rminbyr, 12) - pow(rminbyr, 6))*idr
                fz += fac       ##LJ from the wall

            for j in range(Np):
                dx = r[i   ] - r[j   ]
                dy = r[i+Np] - r[j+Np]
                dz = r[i+xx] - r[j+xx]
                dr = sqrt(dx*dx + dy*dy + dz*dz)
                if i != j and dr < ljr:
                    idr     = 1.0/dr
                    rminbyr = ljr*idr
                    fac   = lje*(pow(rminbyr, 12) - pow(rminbyr, 6))*idr*idr

                    fx += fac*dx
                    fy += fac*dy
                    fz += fac*dz

            F[i]    += fx
            F[i+Np] += fy
            F[i+xx] += fz
        return

    cpdef softSpringWall(self, double [:] F, double [:] r, double pk=0.0100, double prmin=3, double prmax=4,
                         double wlje= 0.001, double wljr = 1.5):
        '''
        lj potential fron wall to particles and spring between particles
        F = -k(r-rmin)
        
        ...
        
        Parameters
        ----------
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
        pk: float
            Strength of harmonic potential between particles
        prmin: float
            Minimum of harmonic potential
        prmax: float
            Cutoff distance of harmonic potential
        
        wlje: float
            Strength of the LJ from wall
        wljr: float
            Range of the LJ from wall

        '''
        cdef int Np=self.Np, i, j, xx=2*Np
        cdef double dx, dy, dz, dr, idr, rminbyr, fac, fx, fy, fz, hh, facss,abshh,sgnhh

        for i in prange(Np, nogil=True):
            fx = 0.0; fy = 0.0; fz = 0.0;
            hh = r[i+xx]
            abshh=sqrt(hh*hh)
            sgnhh=hh/abshh
            idr = 1/hh
            if abshh<wljr:
                  rminbyr = 1.5*idr
                  fac   = 0.1*(pow(rminbyr, 12) - pow(rminbyr, 6))*idr
                  fz+=fac
                # rminbyr = wljr*idr
                # fac   = wlje*(pow(rminbyr, 12) - pow(rminbyr, 6))*idr
                # fz += fac      ##LJ from the wall

            for j in range(Np):
                dx = r[i   ] - r[j   ]
                dy = r[i+Np] - r[j+Np]
                dz = r[i+xx] - r[j+xx]
                dr = sqrt(dx*dx + dy*dy + dz*dz)
                if i != j and dr < prmax:
                    idr     = 1.0/dr
                    fac   = pk*(prmin-dr)

                    fx += fac*dx*idr
                    fy += fac*dy*idr
                    fz += fac*dz*idr

            F[i]    += fx
            F[i+Np] += fy
            F[i+xx] += fz
        return
        
    cpdef softSpringLJWall(self, double [:] F, double [:] r, double pk=0.0100, double prmin=3, double prmax=4,
                         double lje = 0.001, double ljr = 3, double wlje= 0.001, double wljr = 1.5):
        '''
        lj potential fron wall to particles and spring and lj between particles
        F = -k(r-rmin)
        lj stabilises numerical solver when particles get close
        
        ...
        
        Parameters
        ----------
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
        pk: float
            Strength of harmonic potential between particles
        prmin: float
            Minimum of harmonic potential
        prmax: float
            Cutoff distance of harmonic potential
        lje: LJ strength between particles
        ljr: LJ min between particles
        
        wlje: float
            Strength of the LJ from wall
        wljr: float
            Range of the LJ from wall
        '''
        cdef int Np=self.Np, i, j, xx=2*Np
        cdef double dx, dy, dz, dr, idr, rminbyr, fac, fx, fy, fz, hh, facss,abshh,sgnhh

        for i in prange(Np, nogil=True):
            fx = 0.0; fy = 0.0; fz = 0.0;
            hh = r[i+xx]
            abshh=sqrt(hh*hh)
            sgnhh=hh/abshh
            idr = 1/hh
            if abshh<wljr:
                  rminbyr = 1.5*idr
                  fac   = 0.1*(pow(rminbyr, 12) - pow(rminbyr, 6))*idr
                  fz+=fac
                # rminbyr = wljr*idr
                # fac   = wlje*(pow(rminbyr, 12) - pow(rminbyr, 6))*idr
                # fz += fac      ##LJ from the wall

            for j in range(Np):
                dx = r[i   ] - r[j   ]
                dy = r[i+Np] - r[j+Np]
                dz = r[i+xx] - r[j+xx]
                dr = sqrt(dx*dx + dy*dy + dz*dz)
                if i != j and dr < prmax:
                    idr     = 1.0/dr
                    fac   = pk*(prmin-dr)
                    fx += fac*dx*idr
                    fy += fac*dy*idr
                    fz += fac*dz*idr
                    if dr<ljr:
                        idr     = 1.0/dr
                        rminbyr = ljr*idr
                        fac   = lje*(pow(rminbyr, 12) - pow(rminbyr, 6))*idr*idr
                        fx += fac*dx
                        fy += fac*dy
                        fz += fac*dz

                    

            F[i]    += fx
            F[i+Np] += fy
            F[i+xx] += fz
        return
        

    cpdef harmonicRepulsionPPPW(self, double [:] F, double [:] r, double partE=10, double partR=5, double wallE=2, double wallR=5.0):
        cdef int Np = self.Np, i, j, xx = 2*Np
        cdef double dx, dy, dz, dr, idr, rminbyr, fac, fx, fy, fz, hh

        for i in prange(Np, nogil=True):
            fx = 0.0; fy = 0.0; fz = 0.0;
            hh = r[i+xx]
            if hh<wallR:
                fz += wallE*(wallR-hh)

            for j in range(Np):
                dx = r[i   ] - r[j   ]
                dy = r[i+Np] - r[j+Np]
                dz = r[i+xx] - r[j+xx]
                dr = sqrt(dx*dx + dy*dy + dz*dz)
                if i != j and dr < partR:
                    idr     = 1.0/dr
                    fac = partE*(partR - dr)*idr
                    fx += fac*dx
                    fy += fac*dy
                    fz += fac*dz

            F[i]    += fx
            F[i+Np] += fy
            F[i+xx] += fz
        return


    cpdef lennardJonesXWall(self, double [:] F, double [:] r, double wlje=0.12, double wljr=3.0):
        """
        The standard Lennard-Jones potential truncated at the minimum (aslo called WCA potential)
        
            We choose \phi(r) = lje/12 (rr^12 - 2*rr^6 ) + lje/12,  as the standard WCA potential.
            ljr: minimum of the LJ potential and rr=ljr/r.
            This force is only in z-direction due to wall at x=0

        ...

        Parameters
        ----------
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
        lje: float
            Strength of the LJ 
        ljr: float
            Range of the LJ 


        """

        cdef int Np=self.Np, i, j, xx=2*Np
        cdef double dx, dy, dz, dr, idr, rminbyr, fac, fx, fy, fz, hh

        for i in prange(Np, nogil=True):
            fx = 0.0; fy = 0.0; fz = 0.0;
            hh = r[i]+11
            if hh<wljr:
                idr = 1/hh
                rminbyr = wljr*idr
                fac   = wlje*(pow(rminbyr, 12) - pow(rminbyr, 6))*idr
                fx += fac       ##LJ from the wall

            F[i]    += fx
        return
        
    cpdef staticlennardJones(self, double [:] F, double [:] r, double [:] rS, 
                           double lje=0.0100, double ljr=3, double a=1):
        '''
        non-dynamical static particles useful for simulating infinite crystal
        
           ...

        Parameters
        ----------
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
        rS: float
            positions of non-dynamic particles
    
        '''
        cdef int Np=self.Np, i, j, xx=2*Np, Ns = len(rS)/3
        cdef double dx, dy, dz, dr, idr, fac=3*a/4, spring, fx, fy, fz,rminbyr
        for i in prange(Np, nogil=True):
            fx=0; fy=0; fz=0;
            for j in range(Ns):
                dx = r[i   ] - rS[j   ]
                dy = r[i+Np] - rS[j+Ns]
                dz = r[i+xx] - rS[j+Ns*2]
                dr = sqrt(dx*dx + dy*dy + dz*dz)
                if dr < ljr:
                    idr     = 1.0/dr
                    rminbyr = ljr*idr
                    fac   = lje*(pow(rminbyr, 12) - pow(rminbyr, 6))*idr*idr

                    fx += fac*dx
                    fy += fac*dy
                    fz += fac*dz
                    
                
            F[i]    += fx
            F[i+Np] += fy
            F[i+xx] += fz
        return
        
    
    cpdef staticHarmonic(self, double [:] F, double [:] r, double [:] rS, 
                           double pk=0.0100, double prmin=3, double prmax=4,double a=1):
        '''
        non-dynamical static particles useful for simulating infinite crystal
        
           ...

        Parameters
        ----------
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
        rS: float
            positions of non-dynamic particles
    
        '''
        cdef int Np=self.Np, i, j, xx=2*Np, Ns = len(rS)/3
        cdef double dx, dy, dz, dr, idr, fac=3*a/4, spring, fx, fy, fz
        for i in prange(Np, nogil=True):
            fx=0; fy=0; fz=0;
            for j in range(Ns):
                dx = r[i   ] - rS[j   ]
                dy = r[i+Np] - rS[j+Ns]
                dz = r[i+xx] - rS[j+Ns*2]
                dr = sqrt(dx*dx + dy*dy + dz*dz)
                if dr < prmax:
                    '''soft spring repulsion to keep particles away'''
                    idr=1/dr
                    spring= pk*(prmin-dr)
                    fx += spring*dx*idr
                    fy += spring*dy*idr
                    fz += spring*dz*idr
                    
                
            F[i]    += fx
            F[i+Np] += fy
            F[i+xx] += fz
        return


    cpdef harmonicConfinement(self, double [:] F, double [:] r, double cn):
        '''
        Forces on colloids in a harmonic trap
        '''
        cdef int Np = self.Np, i
        for i in prange(3*Np, nogil=True):
            F[i] -= cn*r[i]
        return


    cpdef opticalConfinement(self, double [:] F, double [:] r, double [:] r0, double [:] k):
        """
        Force on colloids in optical traps of varying stiffnesses
        
        ...

        Parameters
        ----------
        r: np.array
            An array of positions
            An array of size 3*Np,
        r0: np.array
            An array of trap centers
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
        k: float 
            Stiffness of the trap
        """

        cdef int Np = self.Np, i, i1, xx = 2*Np
        for i in prange(Np, nogil=True):
            F[i  ]  -= k[i]*(r[i]    - r0[i]  )
            F[i+Np] -= k[i]*(r[i+Np] - r0[i+Np])
            F[i+xx] -= k[i]*(r[i+xx] - r0[i+xx])
        return


    cpdef sedimentation(self, double [:] F, double g):
        """
        Force on colloids in sedimentation 
        
        ...

        Parameters
        ----------
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
        g: float 
            Gravity 
        """
        cdef int Np = self.Np, i, xx = 2*Np
        for i in prange(Np, nogil=True):
#            F[i   ] +=  0
#            F[i+Np] +=  0
            F[i+xx] +=  g
        return


    cpdef membraneConfinement(self, double [:] F, double [:] r, double cn, double r0):
        """
        Force on colloids in membraneSurface 
        
        ...

        Parameters
        ----------
        r: np.array
            An array of positions
            An array of size 3*Np,
        r0: np.array
            An array of trap centers
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
        cn: float 
            Stiffness of the trap
        """

        cdef int Np = self.Np, xx = 2*Np
        cdef int i
        cdef double r0byr
        for i in prange(Np, nogil=True):
            r0byr = (1.0*r0)/sqrt( r[i]*r[i] + r[i+Np]*r[i+Np] + r[i+xx]*r[i+xx] )
            F[i]   -= cn*(1 - r0byr)*r[i]
            F[i+Np] -= cn*(1 - r0byr)*r[i+Np]
            F[i+xx] -= cn*(1 - r0byr)*r[i+xx]
        return


    cpdef membraneBound(self, double [:] F, double [:] r, double cn, double r0):
        """
        Force on colloids in membraneSurface 
        
        ...

        Parameters
        ----------
        r: np.array
            An array of positions
            An array of size 3*Np,
        r0: np.array
            An array of trap centers
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
        cn: float 
            Stiffness of the trap
        """
        cdef int Np = self.Np, xx = 2*Np
        cdef int i
        cdef double r0byr, dist
        for i in prange(Np, nogil=True):
            dist = sqrt( r[i]*r[i] + r[i+Np]*r[i+Np] + r[i+xx]*r[i+xx] )
            if  (dist - r0) > 0.0:
                r0byr = (1.0*r0)/dist
                F[i]   -= cn*(1 - r0byr)*r[i]
                F[i+Np] -= cn*(1 - r0byr)*r[i+Np]
                F[i+xx] -= cn*(1 - r0byr)*r[i+xx]
            else:
                pass
        return


    cpdef spring(self, double [:] F, double [:] r, double bondLength, double springModulus):
        """
        Force on colloids connected by a spring in a single polymer
        
        ...

        Parameters
        ----------
        r: np.array
            An array of positions
        F: np.array
            An array of forces
            An array of size 3*Np,
            An array of size 3*Np,
        bondLength: float
            The size of natural spring 
        springModulus: float 
            Stiffness of the trap
        """

        cdef int Np = self.Np, xx = 2*Np
        cdef int Nf=1, Nm=Np
        cdef:
            int i, ii, ip, jp, kp
            double dx12, dy12, dz12, idr12, dx23, dy23, dz23, idr23,
            double F_bend, F_spring, f1, f3, cos1

        # Spring Force 
        for ii in prange(Nf,nogil=True):
            for i in range(Nm-1):
                ip = Nm*ii + i
                jp = Nm*ii + i + 1
                dx12 = r[ip]    - r[jp]
                dy12 = r[ip+Np] - r[jp+Np]
                dz12 = r[ip+xx] - r[jp+xx] #
                F_spring = springModulus*(bondLength/sqrt(dx12*dx12+dy12*dy12+dz12*dz12) - 1.0) # Scalar part of spring force

                F[ip]      += F_spring*dx12
                F[ip + Np] += F_spring*dy12
                F[ip + xx] += F_spring*dz12
                F[jp]      -= F_spring*dx12
                F[jp + Np] -= F_spring*dy12
                F[jp + xx] -= F_spring*dz12 
        return


    cpdef multipolymers(self, int Nf, double [:] F, double [:] r, double bondLength, double springModulus, double bendModulus, double twistModulus):
        """
        Force on colloids in many polymers connected by a spring 
        
        ...

        Parameters
        ----------
        r: np.array
            An array of positions
        F: np.array
            An array of forces
            An array of size 3*Np,
            An array of size 3*Np,
        bondLength: float
            The size of natural spring 
        springModulus: float 
            Stiffness of the trap
        """
        cdef int Np = self.Np, xx = 2*Np
        cdef:
            int i, ii, ip, jp, kp
            int Nm = Np/Nf # int/int problem might happen
            double dx12, dy12, dz12, idr12, dx23, dy23, dz23, idr23,
            double F_bend, F_spring, f1, f3, cos1

        # Spring Force
        for ii in prange(Nf,nogil=True):
            for i in range(Nm-1):
                ip = Nm*ii + i
                jp = Nm*ii + i + 1
                dx12 = r[ip]    - r[jp]
                dy12 = r[ip+Np] - r[jp+Np]
                dz12 = r[ip+xx] - r[jp+xx] #
                F_spring = springModulus*(bondLength/sqrt(dx12*dx12+dy12*dy12+dz12*dz12) - 1.0) # Scalar part of spring force

                F[ip]      += F_spring*dx12
                F[ip + Np] += F_spring*dy12
                F[ip + xx] += F_spring*dz12
                F[jp]      -= F_spring*dx12
                F[jp + Np] -= F_spring*dy12
                F[jp + xx] -= F_spring*dz12

        # Bending Force
        for ii in prange(Nf,nogil=True):
            for i in range(Nm-2):
                ip = (Nm*ii+i)
                jp = (Nm*ii+i+1)
                kp = (Nm*ii+i+2)
                dx12 = r[ip]   - r[jp]
                dy12 = r[ip+Np] - r[jp+Np]
                dz12 = r[ip+xx] - r[jp+xx] #
                idr12 = 1.0/sqrt( dx12*dx12 + dy12*dy12 + dz12*dz12 )

                dx23 = r[jp]   - r[kp]
                dy23 = r[jp+Np] - r[kp+Np]
                dz23 = r[jp+xx] - r[kp+xx] #
                idr23 = 1.0/sqrt( dx23*dx23 + dy23*dy23 + dz23*dz23 )

                cos1 = (dx12*dx23 + dy12*dy23 + dz12*dz23)
                F_bend = bendModulus*idr12*idr23/bondLength

                f1 = F_bend*( dx23 - dx12*cos1*idr12*idr12)
                f3 = F_bend*(-dx12 + dx23*cos1*idr23*idr23)

                F[ip] += f1
                F[jp] -= f1+f3
                F[kp] += f3

                f1 = F_bend*( dy23 - dy12*cos1*idr12*idr12)
                f3 = F_bend*(-dy12 + dy23*cos1*idr23*idr23)

                F[ip+Np] += f1
                F[jp+Np] -= f1+f3
                F[kp+Np] += f3

                f1 = F_bend*( dz23 - dz12*cos1*idr12*idr12)
                f3 = F_bend*(-dz12 + dz23*cos1*idr23*idr23)

                F[ip+xx] += f1
                F[jp+xx] -= f1+f3
                F[kp+xx] += f3
        return

    cpdef multiRingpolymers(self, int Nf, double [:] F, double [:] r, double bondLength, double springModulus, double bendModulus, double twistModulus):
        """
        Force on colloids connected by a spring in a ring polymer
        
        ...

        Parameters
        ----------
        r: np.array
            An array of positions
        F: np.array
            An array of forces
            An array of size 3*Np,
            An array of size 3*Np,
        bondLength: float
            The size of natural spring 
        springModulus: float 
            Stiffness of the trap
        """
        cdef int Np = self.Np, xx = 2*Np
        cdef:
             int i, ii, ip, jp, kp
             int Nm = Np/Nf # int/int problem might happen
             double dx12, dy12, dz12, idr12, dx23, dy23, dz23, idr23,
             double F_bend, F_spring, f1, f3, cos1

        # Spring Force
        for ii in prange(Nf, nogil=True):
            for i in range(Nm):
                ip = (ii*Nm + i)
                jp = (ii*Nm +((i+1) % Nm)) #NOTE : Nm or Nm-1
                dx12 = r[ip]   - r[jp]
                dy12 = r[ip+Np] - r[jp+Np]
                dz12 = r[ip+xx] - r[jp+xx]
                F_spring = springModulus*(bondLength/sqrt(dx12*dx12+dy12*dy12+dz12*dz12) - 1.0) # Scalar part of spring force

                F[ip]     += F_spring*dx12
                F[ip + Np] += F_spring*dy12
                F[ip + xx] += F_spring*dz12
                F[jp]     -= F_spring*dx12
                F[jp + Np] -= F_spring*dy12
                F[jp + xx] -= F_spring*dz12

        # Bending Force
        for ii in prange(Nf,nogil=True):
            for i in range(Nm):
                ip = ( ii*Nm + i  )
                jp = ( ii*Nm +((i+1) % Nm))
                kp = ( ii*Nm +((i+2) % Nm))
                dx12 = r[ip]      - r[jp]
                dy12 = r[ip+Np]   - r[jp+Np]
                dz12 = r[ip+xx]   - r[jp+xx] #
                idr12 = 1.0/sqrt( dx12*dx12 + dy12*dy12 + dz12*dz12 )

                dx23 = r[jp]   - r[kp]
                dy23 = r[jp+Np] - r[kp+Np]
                dz23 = r[jp+xx] - r[kp+xx] #
                idr23 = 1.0/sqrt( dx23*dx23 + dy23*dy23 + dz23*dz23 )

                cos1 = (dx12*dx23 + dy12*dy23 + dz12*dz23)
                F_bend = bendModulus*idr12*idr23/bondLength

                f1 = F_bend*( dx23 - dx12*cos1*idr12*idr12)
                f3 = F_bend*(-dx12 + dx23*cos1*idr23*idr23)

                F[ip] += f1
                F[jp] -= f1+f3
                F[kp] += f3

                f1 = F_bend*( dy23 - dy12*cos1*idr12*idr12)
                f3 = F_bend*(-dy12 + dy23*cos1*idr23*idr23)

                F[ip+Np] += f1
                F[jp+Np] -= f1+f3
                F[kp+Np] += f3

                f1 = F_bend*( dz23 - dz12*cos1*idr12*idr12)
                f3 = F_bend*(-dz12 + dz23*cos1*idr23*idr23)

                F[ip+xx] += f1
                F[jp+xx] -= f1+f3
                F[kp+xx] += f3
        return


    cpdef membraneSurface(self, int Nmx, int Nmy, double [:] F, double [:] r, double bondLength, double springModulus, double bendModulus):
        """
        Force on colloids connected as a membrane 
        
        ...

        Parameters
        ----------
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
        bondLength: float
            The size of natural spring 
        springModulus: float 
            Stiffness of the trap
        bendModulus: float 
            Bending cost
        """


        cdef int Np = self.Np, xx = 2*Np
        cdef:
            int i1, i2, ip, jp, kp
            double dx12, dy12, dz12, idr12, dx23, dy23, dz23, idr23, cos1, f1, f3
            double F_spring, F_bend
            # pointers in the Position
            double * ri = & r[0]
            double * rj = & r[0]
            double * rk = & r[0]
            # pointers in the Force
            double * Fi = & F[0]
            double * Fj = & F[0]
            double * Fk = & F[0]

        # Spring Force in x direction : first loop in y
        # This loop should be faster compare to the  next for
        for i2 in prange(Nmy,nogil=True):
            for i1 in range(Nmx-1):
                ip = Nmx*i2 + i1
                jp = Nmx*i2 + (i1 + 1)
                dx12 = ri[ip]   - rj[jp]
                dy12 = ri[ip+Np] - rj[jp+Np]
                dz12 = ri[ip+xx] - rj[jp+xx]
                F_spring = springModulus*(bondLength/sqrt(dx12*dx12+dy12*dy12+dz12*dz12) - 1.0) # Scalar part of spring force

                Fi[ip]     += F_spring*dx12
                Fi[ip + Np] += F_spring*dy12
                Fi[ip + xx] += F_spring*dz12
                Fj[jp]     -= F_spring*dx12
                Fj[jp + Np] -= F_spring*dy12
                Fj[jp + xx] -= F_spring*dz12

        # Spring Force in y direction : first loop in x
        # this should be a bit slow as pointer jumps 3*Nmx in every inner loop step
        for i1 in prange(Nmx,nogil=True):
            for i2 in range(Nmy-1):
                ip = Nmx*i2 + i1
                jp = Nmx*(i2+1) + i1
                dx12 = ri[ip]   - rj[jp]
                dy12 = ri[ip+Np] - rj[jp+Np]
                dz12 = ri[ip+xx] - rj[jp+xx]
                F_spring = springModulus*(bondLength/sqrt(dx12*dx12+dy12*dy12+dz12*dz12) - 1.0)

                Fi[ip]     += F_spring*dx12
                Fi[ip + Np] += F_spring*dy12
                Fi[ip + xx] += F_spring*dz12
                Fj[jp]     -= F_spring*dx12
                Fj[jp + Np] -= F_spring*dy12
                Fj[jp + xx] -= F_spring*dz12

        # Bending Force
        for i2 in prange(Nmy,nogil=True):
            for i1 in range(Nmx-2):
                ip = Nmx*i2 + i1
                jp = Nmx*i2 + (i1+1)
                kp = Nmx*i2 + (i1+2)

                dx12 = ri[ip]   - rj[jp]
                dy12 = ri[ip+Np] - rj[jp+Np]
                dz12 = ri[ip+xx] - rj[jp+xx] #
                idr12 = 1.0/sqrt( dx12*dx12 + dy12*dy12 + dz12*dz12 )

                dx23 = rj[jp]   - rk[kp]
                dy23 = rj[jp+Np] - rk[kp+Np]
                dz23 = rj[jp+xx] - rk[kp+xx] #
                idr23 = 1.0/sqrt( dx23*dx23 + dy23*dy23 + dz23*dz23 )

                cos1 = (dx12*dx23 + dy12*dy23 + dz12*dz23)
                F_bend = bendModulus*idr12*idr23/bondLength

                f1 = F_bend*( dx23 - dx12*cos1*idr12*idr12)
                f3 = F_bend*(-dx12 + dx23*cos1*idr23*idr23)

                Fi[ip] += f1
                Fj[jp] -= f1+f3
                Fk[kp] += f3

                f1 = F_bend*( dy23 - dy12*cos1*idr12*idr12)
                f3 = F_bend*(-dy12 + dy23*cos1*idr23*idr23)

                Fi[ip+Np] += f1
                Fj[jp+Np] -= f1+f3
                Fk[kp+Np] += f3

                f1 = F_bend*( dz23 - dz12*cos1*idr12*idr12)
                f3 = F_bend*(-dz12 + dz23*cos1*idr23*idr23)

                Fi[ip+xx] += f1
                Fj[jp+xx] -= f1+f3
                Fk[kp+xx] += f3
        for i1 in prange(Nmx,nogil=True):
            for i2 in range(Nmy-2):
                ip = Nmx*i2     + i1
                jp = Nmx*(i2+1) + i1
                kp = Nmx*(i2+2) + i1

                dx12 = ri[ip]   - rj[jp]
                dy12 = ri[ip+Np] - rj[jp+Np]
                dz12 = ri[ip+xx] - rj[jp+xx] #
                idr12 = 1.0/sqrt( dx12*dx12 + dy12*dy12 + dz12*dz12 )

                dx23 = rj[jp]   - rk[kp]
                dy23 = rj[jp+Np] - rk[kp+Np]
                dz23 = rj[jp+xx] - rk[kp+xx] #
                idr23 = 1.0/sqrt( dx23*dx23 + dy23*dy23 + dz23*dz23 )

                cos1 = (dx12*dx23 + dy12*dy23 + dz12*dz23)
                F_bend = bendModulus*idr12*idr23/bondLength

                f1 = F_bend*( dx23 - dx12*cos1*idr12*idr12)
                f3 = F_bend*(-dx12 + dx23*cos1*idr23*idr23)

                Fi[ip] += f1
                Fj[jp] -= f1+f3
                Fk[kp] += f3

                f1 = F_bend*( dy23 - dy12*cos1*idr12*idr12)
                f3 = F_bend*(-dy12 + dy23*cos1*idr23*idr23)

                Fi[ip+Np] += f1
                Fj[jp+Np] -= f1+f3
                Fk[kp+Np] += f3

                f1 = F_bend*( dz23 - dz12*cos1*idr12*idr12)
                f3 = F_bend*(-dz12 + dz23*cos1*idr23*idr23)

                Fi[ip+xx] += f1
                Fj[jp+xx] -= f1+f3
                Fk[kp+xx] += f3
        return




@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
@cython.wraparound(False)
cdef class Torques:
    """
    Computes torques in a system of colloidal particles

    Methods in the Torques class take input,
        * arrays of positions, Torques 
        * parameters for a given potential
    
    The array of torques is then update by each method. 

    ...

    Parameters
    ----------
    particles: int
        Number of particles (Np)


    """
    def __init__(self, particles=1):
        self.Np = particles 


    cpdef bottomHeaviness(self, double [:] T, double [:] p, double bh=1.0):
        """
        Torque due to bottom-heaviness
        
        ...

        Parameters
        ----------
        p: np.array
            An array of Orientations
            An array of size 3*Np,
        F: np.array
            An array of Torques
            An array of size 3*Np,
        bh: float 
            bottomHeaviness 
        """
        cdef int Np = self.Np, i, xx = 2*Np
        for i in prange(Np, nogil=True):
            T[i   ] +=  -bh*p[i+Np]  # torque is bh\times p
            T[i+Np] +=  bh*p[i]
            T[i+xx] +=  0
        return




cimport cython
from libc.math cimport sqrt, exp, pow, erfc, sin, cos
from cython.parallel import prange
import numpy as np
cimport numpy as np
cdef double PI = 3.14159265359

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
cdef class Rbm:
       
    cdef readonly int Nx, Ny, Nz, Np
    cdef readonly np.ndarray fkx, fky, fkz, vkx, vky, vkz, fk0, fx0
    cdef readonly double Lx, Ly, Lz, a, facx, facy, facz, eta

    cpdef mobilityTT(self, double [:] v, double [:] r, double [:] F, double H)
    
    
    cpdef propulsionT2s(self, double [:] v, double [:] r, double [:] S, double H)
    
    
    cpdef propulsionT3t(self, double [:] v, double [:] r, double [:] D, double H)

    
@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
cdef class Flow:
    cdef double a, eta
    cdef int Np
    cdef int Nt

    cpdef flowField1s(self, double [:] vv, double [:] rt, double [:] r, double [:] F, double H)

    cpdef flowField2s(self, double [:] vv, double [:] rt, double [:] r, double [:] S, double H)

    cpdef flowField3t(self, double [:] vv, double [:] rt, double [:] r, double [:] D, double H)
    



cimport cython
from libc.math cimport sqrt
from cython.parallel import prange
import numpy as np
cimport numpy as np
cdef double PI = 3.14159265359

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
cdef class Rbm:
    cdef double a, eta, mu, muv, mur
    cdef int Np 
    cdef readonly np.ndarray Mobility

    cpdef mobilityTT(self, double [:] v,  double [:] r, double [:] F)
               
   
    cpdef mobilityTR(self,    double [:] v,  double [:] r, double [:] T)
    
    
    cpdef propulsionT2s(self, double [:] v,  double [:] r, double [:] S)


    cpdef propulsionT3t(self, double [:] v,  double [:] r, double [:] D)


    cpdef propulsionT3a(self,   double [:] v,  double [:] r, double [:] V)


    cpdef propulsionT3s(self,   double [:] v,  double [:] r, double [:] G)


    cpdef propulsionT4a(self,   double [:] v,  double [:] r, double [:] M)


    ## Angular velocities


    cpdef mobilityRT(self, double [:] o,  double [:] r, double [:] F)

               
    cpdef mobilityRR(   self, double [:] o,  double [:] r, double [:] T)

    
    cpdef propulsionR2s(self, double [:] o,  double [:] r, double [:] S)
    
    
    cpdef propulsionR3a(  self, double [:] o,  double [:] r, double [:] V)


    cpdef propulsionR3s(  self, double [:] o,  double [:] r, double [:] G)


    cpdef propulsionR4a(  self, double [:] o,  double [:] r, double [:] M) 


    cpdef noiseTT(self, double [:] v, double [:] r)


    cpdef noiseRR(self, double [:] o, double [:] r)



@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
cdef class Flow:
    cdef double a, eta
    cdef int Np
    cdef int Nt

    cpdef flowField1s(self, double [:] vv, double [:] rt, double [:] r, double [:] F)
               
    cpdef flowField2a(   self, double [:] vv, double [:] rt, double [:] r, double [:] T)

    cpdef flowField2s(self, double [:] vv, double [:] rt, double [:] r, double [:] S)


    cpdef flowField3t(self, double [:] vv, double [:] rt, double [:] r, double [:] D)


    cpdef flowField3s(self,   double [:] vv, double [:] rt, double [:] r, double [:] G)


    cpdef flowField3a(self,   double [:] vv, double [:] rt, double [:] r, double [:] V)


    cpdef flowField4a(self,   double [:] vv, double [:] rt, double [:] r, double [:] M)



